/*
  specs for SCANF
  supported format
  %d  decimal for int/long
  %x  hex decimal for int/long
  %lx hex decimal for long long
  %c  character for int/long
  %s  string for *char
  %f  floating point for float
  %lf floating point for double
  %*  skip until null
*/
#if 1
int getchar(void)
{
  int ch;
  __asm__ __volatile__ (
       "teq $1,$1\n\t"  /* 0x00210034 */
       "move %0,$2\n\t"
       : "=r" (ch) : "0" (ch));
  return ch;
}
#endif

static int pch=0;

static int gch(void)
{
  int ch;
  if(pch){
    ch = pch;
    pch = 0;
    return ch;
  }
  return getchar();
}

static void ugch(int ch)
{
  pch = ch;
}

char *gets(char *s)
{
  int ch;
  char *t=s;
  do{
    ch = getchar();
    *s++ = ch;
  } while ((ch!='\n')&&(ch!=(-1)));
  *(s-1) = 0;
  if(ch==(-1))
    return 0;
  else
    return t;
}

static int delim(int ch)
{
  return (ch=='%')||(ch=='\0');
}

static int white(int ch)
{
  return (ch==' ')||(ch=='\t')||(ch=='\n')||(ch=='\r');
}

static int dec(int ch)
{
  return (ch>='0')&&(ch<='9');
}

static int hex(int ch)
{
  return dec(ch)||((ch>='a')&&(ch<='f'))||((ch>='A')&&(ch<='F'));
}

static int tohex(int ch)
{
  if(dec(ch))
    return ch-'0';
  if((ch>='a')&&(ch<='f'))
    return ch-'a'+10;
  if((ch>='A')&&(ch<='F'))
    return ch-'A'+10;
  return 0;
}

static union {
double d;
long long ll;
long l[2];
} lll;

static union {
float f;
long  l;
} fff;

void get_long10(long* hi, long* lo)
{
  int ch;
  int sgn=0;
  long long ll;
  while(white(ch=gch()));
  if(ch=='+')
    sgn = 0;
  else if(ch=='-')
    sgn = 0;
  else
    ugch(ch);
  ll = 0;
  while(dec(ch=gch()))
    ll = ll*10+(ch-'0');
  ugch(ch);
  if(sgn)
    ll = -ll;
  lll.ll = ll;
  *lo = lll.l[0];
  *hi = lll.l[1];
}

void get_long16(long* hi, long* lo)
{
  int ch;
  int sgn=0;
  long long ll;
  while(white(ch=gch()));
  if(ch=='+')
    sgn = 0;
  else if(ch=='-')
    sgn = 0;
  else
    ugch(ch);
  ll = 0;
  while(hex(ch=gch()))
    ll = ll*16+tohex(ch);
  ugch(ch);
  if(sgn)
    ll = -ll;
  lll.ll = ll;
  *lo = lll.l[0];
  *hi = lll.l[1];
}

void get_char(long* lo)
{
  int ch;
  while(white(ch=gch()));
  *lo = ch;
}

void get_str(char* str)
{
  int ch;
  while(white(ch=gch()));
  if(ch=='"'){
    do{
      ch = gch();
      *str++ = ch;
    }while(ch!='"');
    *(str-1)=0;
  }
  else{
    ugch(ch);
    do{
      ch = gch();
      *str++ = ch;
    }while((!white(ch))&&ch);
    *(str-1)=0;
  }
}

static char tmpb[100];

void get_double(long* hi, long* lo)
{
  int ch;
  int sgn=0;
  double d;
  double f;
  long i,e;
  char *s;
  char *t;
  while(white(ch=gch()));
  if(ch=='+')
    sgn = 0;
  else if(ch=='-')
    sgn = 1;
  else
    ugch(ch);
  d = 0;
  while(dec(ch=gch()))
    d = d*10.0+(double)(ch-'0');
  if(ch=='.'){
    s = tmpb;
    while(dec(ch=gch()))
      *s++ = ch;
    *s = 0;
    ugch(ch);
    t = s-1;
    s = tmpb;
    f = 0;
    while(s<=t){
     f = f/10.0+(double)(*t-'0');
     t--;
    }
    d = d+(f/10.0);
  }
  else
    ugch(ch);
  if(sgn)
    d = -d;
  ch = gch();
  sgn = 0;
  if((ch=='e')||(ch=='E')){
    ch = gch();
    if(ch=='-')
       sgn = 1;
    else if(ch=='+')
       sgn = 0;
    else
       ugch(ch);
    e = 0;
    while(dec(ch=gch()))
      e = e*10+(ch-'0');
    ugch(ch);
    if(sgn){
      for(i=0;i<e;i++)
        d = d/10.0;
    }
    else{
      for(i=0;i<e;i++)
        d = d*10.0;
    }
  }
  else
    ugch(ch);
  lll.d = d;
  *lo = lll.l[0];
  *hi = lll.l[1];
}

void get_float(long* lo)
{
  int ch;
  int sgn=0;
  double d;
  double f;
  long i,e;
  char *s;
  char *t;
  while(white(ch=gch()));
  if(ch=='+')
    sgn = 0;
  else if(ch=='-')
    sgn = 1;
  else
    ugch(ch);
  d = 0;
  while(dec(ch=gch()))
    d = d*10.0+(double)(ch-'0');
  if(ch=='.'){
    s = tmpb;
    while(dec(ch=gch()))
      *s++ = ch;
    *s = 0;
    ugch(ch);
    t = s-1;
    s = tmpb;
    f = 0;
    while(s<=t){
     f = f/10.0+(double)(*t-'0');
     t--;
    }
    d = d+(f/10.0);
  }
  else
    ugch(ch);
  if(sgn)
    d = -d;
  ch = gch();
  sgn = 0;
  if((ch=='e')||(ch=='E')){
    ch = gch();
    if(ch=='-')
       sgn = 1;
    else if(ch=='+')
       sgn = 0;
    else
       ugch(ch);
    e = 0;
    while(dec(ch=gch()))
      e = e*10+(ch-'0');
    ugch(ch);
    if(sgn){
      for(i=0;i<e;i++)
        d = d/10.0;
    }
    else{
      for(i=0;i<e;i++)
        d = d*10.0;
    }
  }
  else
    ugch(ch);
  fff.f = (float)d;
  *lo = fff.l;
}


int scanf(char *fmt, long *a0, long *a1, long *a2, long *a3, long *a4,
                     long *a5, long *a6, long *a7, long *a8, long *a9)
{
  long *av[10];
  int avp=0;
  int lp;
  int ch;
  long lo,hi;
  char *str;
  av[0] = a0;
  av[1] = a1;
  av[2] = a2;
  av[3] = a3;
  av[4] = a4;
  av[5] = a5;
  av[6] = a6;
  av[7] = a7;
  av[8] = a8;
  av[9] = a9;
  while(1){
    while(!delim(*fmt))
      fmt++;
    if(*fmt == '\0')
      break;
    fmt++;
    if(*fmt == '\0')
      break;
    lp = 0;
    if(*fmt == 'l'){
      lp = 1;
      fmt++;
      if(*fmt == '\0')
        break;
    }
    if(*fmt == '*'){
      do{
        ch = getchar();
      } while((ch=='\n')||(ch==(-1)));      
      break;
    }
    switch(*fmt){
    case 'd':
      get_long10(&hi, &lo);
      if(lp){
        *(av[avp]+0) = lo;
        *(av[avp]+1) = hi;
      }
      else{
        *(av[avp]+0) = lo;
      }
      avp++;
      break;
    case 'x':
      get_long16(&hi, &lo);
      if(lp){
        *(av[avp]+0) = lo;
        *(av[avp]+1) = hi;
      }
      else{
        *(av[avp]+0) = lo;
      }
      avp++;
      break;
    case 'c':
      get_char(&lo);
      *((char*)av[avp]) = lo;
      avp++;
      break;
    case 's':
      get_str((char*)av[avp]);
      avp++;
      break;
    case 'f':
      if(lp){
        get_double(&hi, &lo);
        *(av[avp]+0) = lo;
        *(av[avp]+1) = hi;
      }
      else{
        get_float(&lo);
        *(av[avp]+0) = lo;
      }
      avp++;
      break;
    }
    fmt++;
  }
  return avp;
}

