/*************************************************/
/* All rights reserved by Akira Nakamori. This   */
/* program must be uesed solely for the purpose  */ 
/* for which it was furnished by Akira Nakamori. */
/*                                               */
/*     Copyright (C) Akira Nakamori 2004         */
/*************************************************/
#include <stdio.h>
#include "define.h"

extern int ty_cpu;
extern int sig_stop;

static ulong ic_data[ICLINES][ICLINE];
static ulong ic_tag [ICLINES];
static int   ic_valid[ICLINES];

static ulong dc_data[DCLINES][DCLINE];
static ulong dc_tag [DCLINES];
static int   dc_valid[DCLINES];
static int   dc_dirty[DCLINES];

int watch_pend = 0;
double ic_refer;
double dc_refer;
double ic_hit;
double dc_hit;

void init_cache(void)
{
  int i;
  for(i=0;i<ICLINES;i++)
	ic_valid[i] = 0;
  for(i=0;i<DCLINES;i++)
	dc_valid[i] = 0;
  ic_refer = 0.0;
  dc_refer = 0.0;
  ic_hit   = 0.0;
  dc_hit   = 0.0;
}

static void check_watch(ulong pa, int type)
{
  ulong a;

  a = mfc0(C0_WatchLo);
  if((watch_pend==0)&&(((a&1)&&(type==WRITE))||((a&2)&&(type==READ)))){
	  watch_pend = (((a^pa)&0xfffffff8L)==0L);
  }
}

ulong ic_read(ulong va, ulong pa)
{
  ulong ptag;
  int   index;
  int   i;
  int   o;

  ic_refer++;
  o = (pa>>2)&(ICLINE-1);
  index = (va/(1<<ICLINE))&(ICLINES-1);
  ptag  = pa & (~((ulong)((1<<ICLINE)-1)));
  if(ic_valid[index]&&(ic_tag[index]==ptag)){ /* hit */
    ic_hit++;
    return ic_data[index][o];
  }
  else{ /* miss & refill */
    pa = ptag;
    for(i=0;i<ICLINE;i++){
      ic_data[index][i] = get_pword(pa);
      pa += 4L;
    }
    ic_valid[index] = 1;
    ic_tag  [index] = ptag;
    return ic_data[index][o];
  }
}

ulong dc_read(ulong va, ulong pa)
{
  ulong ptag;
  ulong pa_wb;
  int   index;
  int   i;
  int   o;

  dc_refer++;
  o = (pa>>2)&(DCLINE-1);
  index = (va/(1<<DCLINE))&(DCLINES-1);
  ptag  = pa & (~((ulong)((1<<DCLINE)-1)));
  if(dc_valid[index]&&(dc_tag[index]==ptag)){ /* hit */
    dc_hit++;
#if 0
    printf("dc_read:%08x %08x\n",pa,dc_data[index][o]);
#endif
    return dc_data[index][o];
  }
  else{ /* miss & writeback & refill */
    if(dc_valid[index]&&dc_dirty[index]){ /* writeback */
      pa_wb = dc_tag[index];
      for(i=0;i<DCLINE;i++){
	put_pword(pa_wb, dc_data[index][i]);
	pa_wb += 4L;
      }
    }
    pa = ptag;
    for(i=0;i<DCLINE;i++){
      dc_data[index][i] = get_pword(pa);
      pa += 4L;
    }
    dc_valid[index] = 1;
    dc_dirty[index] = 0;
    dc_tag  [index] = ptag;
#if 0
    if(dc_data[index][o]==0xfffefffe){
      sig_stop=1;
      printf("dc_read:%08x %08x\n",pa,dc_data[index][o]);
    }
#endif
    return dc_data[index][o];
  }
}

void dc_write(ulong va, ulong pa, ulong data)
{
  ulong ptag;
  ulong pa_wb;
  int   index;
  int   i;
  int   o;

#if 0
    if(data==0xfffefffe){
      sig_stop=1;
      printf("dc_write:%08x %08x\n",pa,data);
    }
#endif
  dc_refer++;
  o = (pa>>2)&(DCLINE-1);
  index = (va/(1<<DCLINE))&(DCLINES-1);
  ptag  = pa & (~((ulong)((1<<DCLINE)-1)));
  if(dc_valid[index]&&(dc_tag[index]==ptag)){ /* hit */
    dc_hit++;
    dc_data[index][o] = data;
    dc_dirty[index] = 1;
  }
  else{ /* miss & writeback & refill */
    if(dc_valid[index]&&dc_dirty[index]){ /* writeback */
      pa_wb = dc_tag[index];
      for(i=0;i<DCLINE;i++){
	put_pword(pa_wb, dc_data[index][i]);
	pa_wb += 4L;
      }
    }
    pa = ptag;
    for(i=0;i<DCLINE;i++){
      dc_data[index][i] = get_pword(pa);
      pa += 4L;
    }
    dc_valid[index] = 1;
    dc_dirty[index] = 1;
    dc_tag  [index] = ptag;
    dc_data[index][o] = data;
  }
}

ulong get_icache(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa;
  int   c;

  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err==Exc_NONE){
    if(c!=2)
      return ic_read(va_lo, pa);
    else
      return get_pword(pa);
  }
  else{
    return 0L; /* No Meaning */
  }
}

ulong get_dcachel(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa;
  int   c;

  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err==Exc_NONE){
	check_watch(pa, READ);
    if(c!=2)
      return dc_read(va_lo, pa);
    else
      return get_pword(pa);
  }
  else{
    return 0L; /* No Meaning */
  }
}

ulong get_dcaches(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa;
  int   c;

  pa = addr_trans(va_hi, va_lo, WRITE, &c, err);
  if(*err==Exc_NONE){
   check_watch(pa, WRITE);
   if(c!=2)
      return dc_read(va_lo, pa);
    else
      return get_pword(pa);
  }
  else{
    return 0L; /* No Meaning */
  }
}

void put_dcache(ulong va_hi, ulong va_lo, ulong data)
{
  ulong pa;
  int   c;
  int   err;

  pa = addr_trans(va_hi, va_lo, WRITE, &c, &err);
  if(err==Exc_NONE){
    if(c!=2)
      dc_write(va_lo, pa, data);
    else
      put_pword(pa, data);
  }
}

void Index_Inv_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa;
  int   c;
  int   index;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  ic_valid[index] = 0;
}

void Index_WB_Inv_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa;
  int   c, v, d;
  int   index;
  int   i;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  v     = dc_valid[index];
  d     = dc_dirty[index];
  if(v&&d){/* dirty -> writeback */
    pa = dc_tag[index];
    for(i=0;i<DCLINE;i++){
      put_pword(pa, dc_data[index][i]);
      pa += 4L;
    }
  }
  dc_valid[index] = 0;
}

void Index_Load_Tag_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, tag;
  int   c;
  int   v;
  int   index;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  tag   = ic_tag[index];
  v     = ic_valid[index];
  if(ty_cpu==R4300){
    tag = ((tag>>4)&0x0fffff00L)|(v<<7);
  }
  else{
    tag = (tag&0xfffffc00L)|(v<<9);
  }
  mtc0(C0_TagLo, tag);
}

void Index_Load_Tag_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, tag;
  int   c;
  int   v, d;
  int   index;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  tag   = dc_tag[index];
  v     = dc_valid[index];
  d     = dc_dirty[index];
  if(ty_cpu==R4300){
    tag = ((tag>>4)&0x0fffff00L)|(v<<7)|(d<<6);
  }
  else{
    tag = (tag&0xfffffc00L)|(v<<9)|(v<<8)|(d<<7);
  }
  mtc0(C0_TagLo, tag);
}

void Index_Store_Tag_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, tag, ptag;
  int   c;
  int   index;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  tag = mfc0(C0_TagLo);
  if(ty_cpu==R4300){
    ptag = ((tag<<4)&0xfffff000L)|(pa&0xffL);
    ptag &= (~((ulong)((1<<ICLINE)-1)));
    ic_tag[index] = ptag;
    ic_valid[index] = ((tag & 0x80)!=0L);
  }
  else{
    ptag = (tag&0xfffffc00L)|(pa&0x3ffL);
    ptag &= (~((ulong)((1<<ICLINE)-1)));
    ic_tag[index] = ptag;
    ic_valid[index] = ((tag & 0x200)!=0L);
  }
}

void Index_Store_Tag_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, tag, ptag;
  int   c;
  int   index;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(ty_cpu==R4300){
    if(*err!=Exc_NONE)
      return;
  }
  else{
    if(*err==Exc_AdEL)
      return;
    *err = Exc_NONE;
  }
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  tag = mfc0(C0_TagLo);
  if(ty_cpu==R4300){
    ptag = ((tag<<4)&0xfffff000L)|(pa&0xfffL);
    ptag &= (~((ulong)((1<<DCLINE)-1)));
    dc_tag[index] = ptag;
    dc_valid[index] = ((tag & 0x80)!=0L);
    dc_dirty[index] = ((tag & 0x40)!=0L);
  }
  else{
    ptag = (tag&0xfffffc00L)|(pa&0x3ffL);
    ptag &= (~((ulong)((1<<DCLINE)-1)));
    dc_tag[index] = ptag;
    dc_valid[index] = ((tag & 0x200)!=0L);
    dc_dirty[index] = ((tag & 0x080)!=0L);
  }
}

void Create_Dirty_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, ptag;
  int   c, v, d;
  int   index;
  int   i;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  v     = dc_valid[index];
  d     = dc_dirty[index];
  if(v&&d){/* dirty -> writeback */
    pa = dc_tag[index];
    for(i=0;i<DCLINE;i++){
      put_pword(pa, dc_data[index][i]);
      pa += 4L;
    }
  }
  ptag = pa & (~((ulong)((1<<DCLINE)-1)));
  dc_valid[index] = 1;
  dc_dirty[index] = 1;
  dc_tag[index]   = ptag;
}

void Hit_Inv_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong ptag;
  ulong pa;
  int   index;
  int   c;
 
  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  ptag = pa & (~((ulong)((1<<ICLINE)-1)));
  if(ic_valid[index]&&(ic_tag[index]==ptag)){ /* hit */
    ic_valid[index] = 0;
  }
}

void Hit_Inv_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong ptag;
  ulong pa;
  int   index;
  int   c;
 
  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  ptag = pa & (~((ulong)((1<<DCLINE)-1)));
  if(dc_valid[index]&&(dc_tag[index]==ptag)){ /* hit */
    dc_valid[index] = 0;
  }
}

void Hit_WB_Inv_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong ptag;
  ulong pa, pa_wb;
  int   index;
  int   i;
  int   c;
 
  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  ptag = pa & (~((ulong)((1<<DCLINE)-1)));
  if(dc_valid[index]&&(dc_tag[index]==ptag)){ /* hit */
    pa_wb = dc_tag[index];
    for(i=0;i<DCLINE;i++){
      put_pword(pa_wb, dc_data[index][i]);
      pa_wb += 4L;
    }
    dc_valid[index] = 0;
  }
}

void Fill_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong ptag;
  ulong pa, pa_wb;
  int   index;
  int   i;
  int   c;
 
  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  ptag = pa & (~((ulong)((1<<ICLINE)-1)));
  pa_wb = ptag;
  for(i=0;i<ICLINE;i++){
    ic_data[index][i] = get_pword(pa_wb);
    pa_wb += 4L;
  }
  ic_valid[index] = 1;
  ic_tag[index] = ptag;
}

void Hit_WB_I(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, ptag, pa_wb;
  int   c;
  int   index;
  int   i;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<ICLINE))&(ICLINES-1);
  ptag = pa & (~((ulong)((1<<ICLINE)-1)));
  if(ic_valid[index]&&(ic_tag[index]==ptag)){ /* hit */
    pa_wb = ic_tag[index];
    for(i=0;i<ICLINE;i++){
      put_pword(pa_wb, ic_data[index][i]);
      pa_wb += 4L;
    }
  }
}

void Hit_WB_D(ulong va_hi, ulong va_lo, int *err)
{
  ulong pa, ptag, pa_wb;
  int   c;
  int   index;
  int   i;

  if(va_lo&0x3L){
    *err = Exc_AdEL;
    set_badva(va_hi, va_lo);
    return;
  }
  pa = addr_trans(va_hi, va_lo, READ, &c, err);
  if(*err!=Exc_NONE)
    return;
  if(c==2)/* uncache */
    return;
  index = (va_lo/(1<<DCLINE))&(DCLINES-1);
  ptag = pa & (~((ulong)((1<<DCLINE)-1)));
  if(dc_valid[index]&&dc_dirty[index]&&(dc_tag[index]==ptag)){ /* hit */
    pa_wb = dc_tag[index];
    for(i=0;i<DCLINE;i++){
      put_pword(pa_wb, dc_data[index][i]);
      pa_wb += 4L;
    }
  }
}


 


