/*************************************************/
/* All rights reserved by Akira Nakamori. This   */
/* program must be uesed solely for the purpose  */ 
/* for which it was furnished by Akira Nakamori. */
/*                                               */
/*     Copyright (C) Akira Nakamori 2004         */
/*************************************************/
#include <stdio.h>
#include "define.h"

void do_save_restore(ulong extend, ulong insn, int mode);

extern int ty_cpu;
extern int ty_llsc;
extern int mips32_en;
extern int no_mips4;
extern int watch_pend;
extern int extend_flag;
extern int ps_en;
extern int m3d_en;
extern int orig_en;
extern int ver2_en;
extern int sig_stop;
ulong brk_inst;
int   except;
int   addr64;
static ulong PCH,PCL;
static ulong PCH_p,PCL_p;

static int dec_op, /* opcode */
           dec_rs, /* source register */
           dec_rt, /* target register */
           dec_rd, /* destination register */
           dec_sa, /* shift amount */
           dec_fn, /* function code */
           dec_imm;/* immediate value */
static int   LLbit = 0;
static int   delay_slot = 0;
int   br_inst = 0;
int   br_cond = 0;
int   br_likely = 0;
int   no_ds_br = 0;
int   inst64_en;
static int   mips4_en;
static int   cp0_en;
static int   cp1_en;
ulong target_hi, target_lo;
static ulong src_hi, src_lo;
static ulong tar_hi, tar_lo;
static ulong dst_hi, dst_lo;

static ulong GR[32][2];
static ulong HIHI,HILO;
static ulong LOHI,LOLO;

static ulong FR[32][2];
static ulong FCR31;
static int   fr_bit;

void init_cpu(void)
{
  delay_slot = 0;
  br_inst = 0;
  br_cond = 0;
  br_likely = 0;
  no_ds_br = 0;
}

void set_fcr31(ulong lo)
{
  FCR31 = lo&0xff83ffffL;
}

void get_fr64(int no, ulong *hi, ulong *lo)
{
  if(fr_bit){
    no &= 0x1f;
    *lo = FR[no][0];
    *hi = FR[no][1];
  }
  else{
    no &= 0x1e;
    *lo = FR[no][0];
    *hi = FR[no][1];
  }
}

ulong get_fr32(int no)
{
  int i;

  if(fr_bit){
    no &= 0x1f;
    return FR[no][0];
  }
  else{
    i  = no&1;
    no &= 0x1e;
    return FR[no][i];
  }
}

void set_fr64(int no, ulong hi, ulong lo)
{
  if(fr_bit){
    no &= 0x1f;
    FR[no][0] = lo;
    FR[no][1] = hi;
  }
  else{
    no&= 0x1e;
    FR[no][0] = lo;
    FR[no][1] = hi;
  }
}

void set_fr32(int no, ulong d)
{
  int i;

  if(fr_bit){
    no &= 0x1f;
    FR[no][0] = d;
    FR[no][1] = 0;
  }
  else{
    i  = no&1;
    no &= 0x1e;
    FR[no][i] = d;
  }
}

void set_pc(ulong hi, ulong lo)
{
  PCH = hi;
  PCL = lo;
}

void get_pc(ulong *hi, ulong *lo)
{
  *hi = PCH;
  *lo = PCL;
}

void get_gr(int no, ulong *hi, ulong *lo)
{
  no &= 0x1f;
  if(no==0){
    *hi = 0L;
    *lo = 0L;
  }
  else{
    *hi = GR[no][0];
    *lo = GR[no][1];
  }
}

void set_gr(int no, ulong hi, ulong lo)
{
#if 0
  printf("set_gr(%d) %08x %08x\n",no,hi,lo);
#endif
  no &= 0x1f;
  if(no){
    GR[no][0] = hi;
    GR[no][1] = lo;
  }
}

void get_hi(ulong *hi, ulong *lo)
{
  *hi = HIHI;
  *lo = HILO;
}

void set_hi(ulong hi, ulong lo)
{
  HIHI = hi;
  HILO = lo;
}

void get_lo(ulong *hi, ulong *lo)
{
  *hi = LOHI;
  *lo = LOLO;
}

void set_lo(ulong hi, ulong lo)
{
  LOHI = hi;
  LOLO = lo;
}

ulong sign_extend(ulong a)
{
  if(a&0x80000000L){
    return 0xffffffffL;
  }
  else{
    return 0L;
  }
}

static void ea_gen(ulong base_hi, ulong base_lo, int dec_imm,
		   ulong *ea_hi, ulong *ea_lo)
{
  ulong immd_hi, immd_lo;
  int   cy;

  if(dec_imm&0x8000){
    immd_hi = 0xffffffffL;
    immd_lo = 0xffff0000L|dec_imm;
  }
  else{
    immd_hi = 0L;
    immd_lo = dec_imm;
  }
  IADD64(base_hi, base_lo, immd_hi, immd_lo, ea_hi, ea_lo, &cy);
}

#if 0
static void saturate(int us, ulong sh, ulong sl,
                             ulong th, ulong tl,
		             ulong *oh, ulong *ol)
{
  int   cy, ds;
  ulong dh, dl;

  IADD64(sh, sl, th, tl, &dh, &dl, &cy);

  ds = ((dh&0x80000000L)!=0L);

  if(us&&(dh!=0L)){
    *oh = 0xffffffffL;
    *ol = 0xffffffffL;
  }
  else if(us==0){
    if((!ds)&&(dl>0x7fffffffL)){
      *oh = 0L;
      *ol = 0x7fffffffL;
    }
    else if(ds&&(dl<0x80000000L)){
      *oh = 0xffffffffL;
      *ol = 0x80000000L;
    }
    else{
      *oh = dh;
      *ol = dl;
    }
  }
  else{
    *oh = dh;
    *ol = dl;
  }
}
#else
static void saturate(int us, int sxort,
                     ulong sh, ulong sl, /* MULT RESULT */
                     ulong th, ulong tl, /* HILO */
		     ulong *oh, ulong *ol)
{
  int   cy, ds;
  ulong dh, dl;
  int   of,uf;
  int   ms;
  int   as;

  IADD64(sh, sl, th, tl, &dh, &dl, &cy);
  ds = ((dh&0x80000000L)!=0L);
  ms = (us)? (dh&1) : ((dl&0x80000000L)!=0L);
  as = (us)? 0      : ((th&0x80000000L)!=0L);
  uf = (!ms) & sxort  & as;
  of = ms    &(!sxort)& (!as);

  if(of){
    if(us){
      *oh = 0xffffffffL;
      *ol = 0xffffffffL;
    }
    else{
      *oh = 0L;
      *ol = 0x7fffffffL;
    }
  }
  else if(uf){
    *oh = 0xffffffffL;
    *ol = 0x80000000L;
  }
  else{
    *oh = dh;
    *ol = dl;
  }
}
#endif

static void exec_macc5432(int op, ulong rs, ulong rt, ulong *rd)
{
  ulong hh,hl,lh,ll,th,tl;
  ulong hs,ht;
  int u,h,s,a;
  u = op&1;
  h = ((op&(1<<9))!=0);
  a = ((op&(1<<8))!=0);
  s = ((op&(1<<7))!=0);
  hs = (u)? 0 : sign_extend(rs);
  ht = (u)? 0 : sign_extend(rt);
  UMUL64(hs, rs, ht, rt, &hh, &hl, &lh, &ll);
  if(s){
    INEG64(lh, ll, &th, &tl);
    lh = th;
    ll = tl;
  }
  if(a){
    IADD64(HILO, LOLO, lh, ll, &th, &tl, &a);
    lh = th;
    ll = tl;
  }
  HILO = lh;
  HIHI = sign_extend(lh);
  LOLO = ll;
  LOHI = sign_extend(ll);
  if(h)
    *rd = lh;
  else
    *rd = ll;
}

static void exec_macc(int sat, int hi, int us)
{
  ulong temp1_h, temp1_l;
  ulong temp2_h, temp2_l;
  int   cy;
  int   sxort;

  sxort = ((src_lo&0x80000000L)!=0)^((tar_lo&0x80000000L)!=0);
  switch((sat<<2)|(hi<<1)|us){
  case 0:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l, &cy);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 1:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l, &cy);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 2:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l, &cy);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 3:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l, &cy);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 4:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 5:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 6:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 7:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, HILO, LOLO, &temp2_h, &temp2_l);
    HILO = temp2_h;
    HIHI = sign_extend(temp2_h);
    LOLO = temp2_l;
    LOHI = sign_extend(temp2_l);
    set_gr(dec_rd, HIHI, HILO);
    break;
  }
}

static void exec_dmacc(int sat, int hi, int us)
{
  ulong temp1_h, temp1_l;
  ulong temp2_h, temp2_l;
  int   cy;
  int   sxort;

  sxort = ((src_lo&0x80000000L)!=0)^((tar_lo&0x80000000L)!=0);
  switch((sat<<2)|(hi<<1)|us){
  case 0:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l, &cy);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 1:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l, &cy);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 2:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l, &cy);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 3:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    IADD64(temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l, &cy);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 4:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 5:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 6:
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 7:
    UMUL64(0L, src_lo, 0L, tar_lo,
	   &temp2_h, &temp2_l, &temp1_h, &temp1_l);
    saturate(us,sxort,temp1_h, temp1_l, LOHI, LOLO, &temp2_h, &temp2_l);
    LOHI = temp2_h;
    LOLO = temp2_l;
    set_gr(dec_rd, HIHI, HILO);
    break;
  }
}

static void check_fp_except(void)
{
  int cause;
  int mask;
  if(FCR31&0x20000L){
    except = Exc_FPE;
  }
  else{
    cause = (FCR31>>12)&0x1f;
    mask  = (FCR31>>7)&0x1f;
    if(cause&mask){
      except = Exc_FPE;
    }
  }
}

static void check_set_fr32(int no, ulong d,
      int e, int v, int z, int o, int u, int i, int cmp, int cc)
{
  int cause;
  int mask;
  ulong m,n;
#if 0
  if(e||v||z||o||u){
    ulong sh,sl,th,tl;
    get_fr64(dec_rd,&sh,&sl);
    get_fr64(dec_rt,&th,&tl);
    sig_stop=1;
    printf("%08x evzou=%d%d%d%d%d\n",PCL,e,v,z,o,u);
    printf("fs=%08x %08x ft=%08x %08x\n",sh,sl,th,tl);
  }
#endif
  if(e){
    cause = (v<<4)|(z<<3)|(o<<2)|(u<<1)|i;
    FCR31 = (FCR31&0xff800fffL)|(cause<<12);
    FCR31 |= 0x20000L;
    except = Exc_FPE;
  }
  else{
    cause = (v<<4)|(z<<3)|(o<<2)|(u<<1)|i;
    mask  = (FCR31>>7)&0x1f;
    FCR31 = (FCR31&0xff800fffL)|(cause<<12);
    if(cause&mask){
      except = Exc_FPE;
    }
    else{
      FCR31 |= (cause<<2);
      if(cmp){
        no = (no>>2)&7;
        if(no==0){
          n = (1<<23);
          m = ~n;
        }
        else{
          n = (1<<(no+24));
          m = ~n;
        }
#if 0
        if(cc&&(cause==0)){
#else
        if(cc){
#endif
          FCR31 |= n;
        }
        else{
          FCR31 &= m;
        }
      }
      else{
	no &= 0x1e;
#if 0
	m = sign_extend(d);
#else
	m = 0;
#endif
        set_fr64(no, m, d);
      }
    }
  }
}

static void check_set_fr64(int no, ulong hi, ulong lo,
      int e, int v, int z, int o, int u, int i, int cmp, int cc)
{
  int cause;
  int mask;
  ulong m,n;
#if 0
  if(e||v||z||o||u){
    ulong sh,sl,th,tl;
    get_fr64(dec_rd,&sh,&sl);
    get_fr64(dec_rt,&th,&tl);
    sig_stop=1;
    printf("%08x evzou=%d%d%d%d%d\n",PCL,e,v,z,o,u);
    printf("fs=%08x %08x ft=%08x %08x\n",sh,sl,th,tl);
  }
#endif
  if(e){
    cause = (v<<4)|(z<<3)|(o<<2)|(u<<1)|i;
    FCR31 = (FCR31&0xff800fffL)|(cause<<12);
    FCR31 |= 0x20000L;
    except = Exc_FPE;
  }
  else{
    cause = (v<<4)|(z<<3)|(o<<2)|(u<<1)|i;
    mask  = (FCR31>>7)&0x1f;
    FCR31 = (FCR31&0xff800fffL)|(cause<<12);
    if(cause&mask){
      except = Exc_FPE;
    }
    else{
      FCR31 |= (cause<<2);
      if(cmp){
	if(cmp>1){
	  no = (no>>2)&7;
	  if(no==0){
	    n = (5<<23);
	    m = ~n;
	    n = (((cc>>1)<<2)|(cc&1))<<23;
	  }
	  else{
	    n = (3<<(no+24));
	    m = ~n;
	    n = (cc<<(no+24));
	  }
	  FCR31 &= m;
	  FCR31 |= n;
	}
	else{
	  no = (no>>2)&7;
	  if(no==0){
	    n = (1<<23);
	    m = ~n;
	  }
	  else{
	    n = (1<<(no+24));
	    m = ~n;
	  }
	  if(cc){
	    FCR31 |= n;
	  }
	  else{
	    FCR31 &= m;
	  }
	}
      }
      else{
        set_fr64(no, hi, lo);
      }
    }
  }
}

static ulong fp_neg(ulong x)
{
  int  e;
  long m;
  e = (x>>23)&0xff;
  m = (x & ((1<<23)-1));
  if((e==0xff)&&(m!=0))
    return x;
  return x^0x80000000;
}

static void exec_cop1(void)
{
  ulong base_hi, base_lo;
  ulong immd_hi, immd_lo;
  ulong fsh, fsl, fth, ftl, fdh, fdl, dd;
  int   e, v, z, o, u, i;
  int   e2, v2, z2, o2, u2, i2;
  int   cy, ch, cmp, cc, fs, rm;
  int   cc1,cc2,cc3,ch1,ch2,ch3;

  if(!cp1_en){
    except = Exc_CpU;
    return;
  }
  fs = ((FCR31&((ulong)(1<<24)))!=0L);
  rm = FCR31&3;
  switch(dec_rs){
  case 000:/* MF */
    dst_lo = get_fr32(dec_rd);
    dst_hi = sign_extend(dst_lo);
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 001:/* DMF */
    if(inst64_en){
      get_fr64(dec_rd, &dst_hi, &dst_lo);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 002:/* CF */
    if(dec_rd==0)
      set_gr(dec_rt, 0L, mfc0(C0_PRId));
    if(dec_rd==25)
      set_gr(dec_rt, 0L, ((FCR31>>24)&0xfe)|((FCR31>>23)&1));
    if(dec_rd==26)
      set_gr(dec_rt, 0L, FCR31&0x3f07c);
    if(dec_rd==28)
      set_gr(dec_rt, 0L, FCR31&0xf87);
    if(dec_rd==31)
      set_gr(dec_rt, sign_extend(FCR31), FCR31);
    break;
  case 003:/* MFH */
    if(ver2_en){
      get_fr64(dec_rd, &dst_hi, &dst_lo);
      set_gr(dec_rt, sign_extend(dst_hi), dst_hi);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 004:/* MT */
    if(fr_bit)
      set_fr64(dec_rd, tar_hi, tar_lo);
    else
      set_fr32(dec_rd, tar_lo);
    break;
  case 005:/* DMT */
    if(inst64_en){
      set_fr64(dec_rd, tar_hi, tar_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 006:/* CT */
    if(dec_rd==31){
      if(mips4_en)
	FCR31 = tar_lo&0xff83ffffL;
      else
	FCR31 = tar_lo&0x0183ffffL;
      check_fp_except();
    }
    if(dec_rd==25){
      FCR31 = (FCR31&0x017fffffL)|((tar_lo&0xfe)<<24)|((tar_lo&1)<<23);
    }
    if(dec_rd==26){
      FCR31 = (FCR31&0xfffc0f83L)|(tar_lo&0x3f07c);
      check_fp_except();
    }
    if(dec_rd==28){
      FCR31 = (FCR31&0xfffff078L)|(tar_lo&0x0f87);
      check_fp_except();
    }
    break;
  case 007:/* MTH */
    if(ver2_en){
      get_fr64(dec_rd, &dst_hi, &dst_lo);
      set_fr64(dec_rd, tar_lo, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 010:/* BC */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    cc = (dec_rt>>2)&7;
    if(cc==0)
      ch = ((FCR31&((ulong)(1<<23)))!=0L);
    else
      ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
    br_inst = 4;
    switch(dec_rt&3){
    case 000:/* BCF */
      br_cond    = (ch==0);
      br_likely  = 0;
      break;
    case 001:/* BCT */
      br_cond    = (ch==1);
      br_likely  = 0;
      break;
    case 002:/* BCFL */
      br_cond    = (ch==0);
      br_likely  = 1;
      break;
    case 003:/* BCTL */
      br_cond    = (ch==1);
      br_likely  = 1;
      break;
    default:
      except = Exc_RI;
    }
    break;
  case 011:/* BC1ANY2 */
    if(m3d_en){
      IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
      if(dec_imm&0x8000){
	immd_hi = 0xffffffffL;
	immd_lo = 0xfffc0000L|(dec_imm<<2);
      }
      else{
	immd_hi = 0L;
	immd_lo = dec_imm<<2;
      }
      IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
      cc = (dec_rt>>2)&6;
      cc1 = cc+1;
      if(cc==0)
	ch = ((FCR31&((ulong)(1<<23)))!=0L);
      else
	ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
      ch1 = ((FCR31&((ulong)(1<<(cc1+24))))!=0L);
      br_inst = 4;
      switch(dec_rt&3){
      case 000:/* BCF */
	br_cond    = ((ch==0)||(ch1==0));
	br_likely  = 0;
	break;
      case 001:/* BCT */
	br_cond    = ((ch==1)||(ch1==1));
	br_likely  = 0;
	break;
      default:
	except = Exc_RI;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 012:/* BC1ANY4 */
    if(m3d_en){
      IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
      if(dec_imm&0x8000){
	immd_hi = 0xffffffffL;
	immd_lo = 0xfffc0000L|(dec_imm<<2);
      }
      else{
	immd_hi = 0L;
	immd_lo = dec_imm<<2;
      }
      IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
      cc = (dec_rt>>2)&4;
      cc1 = cc+1;
      cc2 = cc+2;
      cc3 = cc+3;
      if(cc==0)
	ch = ((FCR31&((ulong)(1<<23)))!=0L);
      else
	ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
      ch1 = ((FCR31&((ulong)(1<<(cc1+24))))!=0L);
      ch2 = ((FCR31&((ulong)(1<<(cc2+24))))!=0L);
      ch3 = ((FCR31&((ulong)(1<<(cc3+24))))!=0L);
      br_inst = 4;
      switch(dec_rt&3){
      case 000:/* BCF */
	br_cond    = ((ch==0)||(ch1==0)||(ch2==0)||(ch3==0));
	br_likely  = 0;
	break;
      case 001:/* BCT */
	br_cond    = ((ch==1)||(ch1==1)||(ch2==1)||(ch3==1));
	br_likely  = 0;
	break;
      default:
	except = Exc_RI;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 020:/* S */
    fsl = get_fr32(dec_rd);
    fsh = 0L;
    ftl = get_fr32(dec_rt);
    fth = 0L;
    switch(dec_fn){
    case 046:/* CVT.PS.S */
      if(ps_en){
        FPU(opABS, fmtS, rm, fs, fsh, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
	if(v){
	  fsl = 0x7fbfffff;
	}
        FPU(opABS, fmtS, rm, fs, fth, ftl, fth, ftl,
            &fdh, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
	if(v2){
	  ftl = 0x7fbfffff;
	}
	check_set_fr64(dec_sa, fsl, ftl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
/*	set_fr64(dec_sa, fsl, ftl);*/
      }
      else{
        except = Exc_RI;
      }
      break;
    case opMOV:
#if 0
      set_fr64(dec_sa, 0, fsl);
#else
      get_fr64(dec_rd, &fsh, &fsl);
      set_fr64(dec_sa, fsh, fsl);
#endif
      break;
    case opMOVCF:
      if(mips4_en){
        cc = (dec_rt>>2)&7;
        if(cc==0)
          ch = ((FCR31&((ulong)(1<<23)))!=0L);
        else
          ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
        cc = dec_rt&1;
        if(ch==cc){
#if 0
          set_fr64(dec_sa, 0, fsl);
#else
	  get_fr64(dec_rd, &fsh, &fsl);
	  set_fr64(dec_sa, fsh, fsl);
#endif
	}
      }
      else{
        except = Exc_RI;
      }
      break;
    case opMOVZ:
    case opMOVN:
      if(mips4_en){
        get_gr(dec_rt, &fth, &ftl);
        ch = ((fth|ftl)==0L);
        cc = dec_fn&1;
        if(ch!=cc){
#if 0	
          set_fr32(dec_sa, fsl);
#else
	  get_fr64(dec_rd, &fsh, &fsl);
	  set_fr64(dec_sa, fsh, fsl);
#endif
	}
      }
      else{
        except = Exc_RI;
      }
      break;
    case opROUND_L:
    case opTRUNC_L:
    case opCEIL_L:
    case opFLOOR_L:
    case opCVT_L:
      if(inst64_en){
        FPU(dec_fn, fmtS, rm, fs, fsh, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      }
      else{
        except = Exc_RI;
      }
      break;
    case opROUND_W:
    case opTRUNC_W:
    case opCEIL_W:
    case opFLOOR_W:
    case opCVT_W:
      FPU(dec_fn, fmtS, rm, fs, fsh, fsl, fth, ftl,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, sign_extend(fdl), fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case opCVT_D:
      FPU(dec_fn, fmtS, rm, fs, fsh, fsl, fth, ftl,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case opRECIP:
    case opRSQRT:
      if(mips4_en){
        FPU(dec_fn, fmtS, rm, fs, fsh, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
      }
      else{
        except = Exc_RI;
      }
      break;
    case 034:/* RECIP2 */
      {
	long rh,rl;
	rh = 0x00000000L;
	rl = 0x3f800000L;
	FPUX(070/* NMSUB.S */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
	     &e, &v, &z, &o, &u, &i);
	if(e){
	  v = z = o = u = i = 0;
	}
	check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, 0, 0);
      }
      break;
    case 035:/* RECIP1 */
      FPU(opRECIP, fmtS, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case 036:/* RSQRT1 */
      FPU(opRSQRT, fmtS, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case 037:/* RSQRT2 */
      {
	long rh,rl;
	rh = 0x00000000L;
	rl = 0x3f800000L;
	FPUX(070/* NMSUB.S */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
	     &e, &v, &z, &o, &u, &i);
	if(e||v){
	  e2 = v2 = z2 = o2 = u2 = i2 = 0;
	  if(e){
	    v = z = o = u = i = 0;
	  }
	  else{
	    fdl = 0x7fbfffff;
	  }
	}
	else{
	  fsh = fdh;
	  fsl = fdl;
	  fth = 0x00000000L;
	  ftl = 0x40000000L;
	  FPU(opDIV, fmtS, rm, fs, fsh, fsl, fth, ftl,
	      &fdh, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	  if(e2||v2){
	    if(e2){
	      v2 = z2 = o2 = u2 = i2 = 0;
	    }
	    else{
	      fdl = 0x7fbfffff;
	    }
	  }
	}
	check_set_fr64(dec_sa, 0, fdl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
      }
      break;
    default:
      if(m3d_en&&(dec_fn>=060)&&(dec_sa&1)){
	fsl &= 0x7fffffffL;
	ftl &= 0x7fffffffL;
      }
      FPU(dec_fn, fmtS, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
    }
    break;
  case 021:/* D */
    get_fr64(dec_rd, &fsh, &fsl);
    get_fr64(dec_rt, &fth, &ftl);
    switch(dec_fn){
    case opMOV:
      set_fr64(dec_sa, fsh, fsl);
      break;
    case opMOVCF:
      if(mips4_en){
        cc = (dec_rt>>2)&7;
        if(cc==0)
          ch = ((FCR31&((ulong)(1<<23)))!=0L);
        else
          ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
        cc = dec_rt&1;
        if(ch==cc)
          set_fr64(dec_sa, fsh, fsl);
      }
      else{
        except = Exc_RI;
      }
      break;
    case opMOVZ:
    case opMOVN:
      if(mips4_en){
        get_gr(dec_rt, &fth, &ftl);
        ch = ((fth|ftl)==0L);
        cc = dec_fn&1;
        if(ch!=cc)
          set_fr64(dec_sa, fsh, fsl);
      }
      else{
        except = Exc_RI;
      }
      break;
    case opROUND_L:
    case opTRUNC_L:
    case opCEIL_L:
    case opFLOOR_L:
    case opCVT_L:
      if(inst64_en){
        FPU(dec_fn, fmtD, rm, fs, fsh, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      }
      else{
        except = Exc_RI;
      }
      break;
    case opROUND_W:
    case opTRUNC_W:
    case opCEIL_W:
    case opFLOOR_W:
    case opCVT_W:
      FPU(dec_fn, fmtD, rm, fs, fsh, fsl, fth, ftl,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, sign_extend(fdl), fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case opCVT_S:
      FPU(dec_fn, fmtD, rm, fs, fsh, fsl, fth, ftl,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case opRECIP:
    case opRSQRT:
      if(mips4_en){
        FPU(dec_fn, fmtD, rm, fs, fsh, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
     }
      else{
        except = Exc_RI;
      }
      break;
    case 034:/* RECIP2 */
      {
	long rh,rl;
	rh = 0x3ff00000L;
	rl = 0x00000000L;
	FPUX(071/* NMSUB.D */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
	     &e, &v, &z, &o, &u, &i);
	if(e){
	  v = z = o = u = i = 0;
	}
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, 0, 0);
      }
      break;
    case 035:/* RECIP1 */
      FPU(opRECIP, fmtD, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case 036:/* RSQRT1 */
      FPU(opRSQRT, fmtD, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case 037:/* RSQRT2 */
      {
	long rh,rl;
	rh = 0x3ff00000L;
	rl = 0x00000000L;
	FPUX(071/* NMSUB.D */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
	     &e, &v, &z, &o, &u, &i);
	if(e||v){
	  e2 = v2 = z2 = o2 = u2 = i2 = 0;
	  if(e){
	    v = z = o = u = i = 0;
	  }
	  else{
	    fdh = 0x7ff7ffff;
	    fdl = 0xffffffff;
	  }
	}
	else{
	  fsh = fdh;
	  fsl = fdl;
	  fth = 0x40000000L;
	  ftl = 0x00000000L;
	  FPU(opDIV, fmtD, rm, fs, fsh, fsl, fth, ftl,
	      &fdh, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	  if(e2||v2){
	    if(e2){
	      v2 = z2 = o2 = u2 = i2 = 0;
	    }
	    else{
	      fdh = 0x7ff7ffff;
	      fdl = 0xffffffff;
	    }
	  }
	}
	check_set_fr64(dec_sa, fdh, fdl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
      }
      break;
    default:
      if(m3d_en&&(dec_fn>=060)&&(dec_sa&1)){
	fsh &= 0x7fffffffL;
	fth &= 0x7fffffffL;
      }
      FPU(dec_fn, fmtD, rm, fs, fsh, fsl, fth, ftl,
	  &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
   }
    break;
  case 024:/* W */
    fsl = get_fr32(dec_rd);
    fsh = 0L;
    switch(dec_fn){
    case opCVT_D:
      FPU(dec_fn, fmtW, rm, fs, fsh, fsl, 0L, 0L,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case opCVT_S:
      FPU(dec_fn, fmtW, rm, fs, fsh, fsl, 0L, 0L,
          &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
      break;
    case 046:/* CVT.PS.PW */
      get_fr64(dec_rd, &fsh, &fsl);
      FPU(opCVT_S, fmtW, rm, fs, 0, fsl, fth, ftl,
	  &dd, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
      FPU(opCVT_S, fmtW, rm, fs, 0, fsh, fth, ftl,
	  &dd, &fdh, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
      if(e){
	v = z = o = u = i = 0;
      }
      if(e2){
	v2 = z2 = o2 = u2 = i2 = 0;
      }
      check_set_fr64(dec_sa, fdh, fdl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
      break;
    default:
      check_set_fr32(0, 0L, 1, 0, 0, 0, 0, 0, 0, 0);
    }
    break;
  case 025:/* L */
    if(inst64_en){
      get_fr64(dec_rd, &fsh, &fsl);
      switch(dec_fn){
      case opCVT_D:
        FPU(dec_fn, fmtL, rm, fs, fsh, fsl, 0L, 0L,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
        break;
      case opCVT_S:
        FPU(dec_fn, fmtL, rm, fs, fsh, fsl, 0L, 0L,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, 0, fdl, e, v, z, o, u, i, cmp, cc);
        break;
      default:
        check_set_fr32(0, 0L, 1, 0, 0, 0, 0, 0, 0, 0);
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 026:/* PS */
    if(ps_en){
      get_fr64(dec_rd, &fsh, &fsl);
      get_fr64(dec_rt, &fth, &ftl);
      get_fr64(dec_sa, &fdh, &fdl);
      switch(dec_fn){
      case opMOV:
	set_fr64(dec_sa, fsh, fsl);
	break;
      case opMOVCF:
	if(mips4_en){
	  int cc2;
	  int ch2;
	  cc = (dec_rt>>2)&7;
	  cc2 = (cc+1)&7;
	  if(cc==0)
	    ch = ((FCR31&((ulong)(1<<23)))!=0L);
	  else
	    ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
	  ch2 = ((FCR31&((ulong)(1<<(cc2+24))))!=0L);
	  cc = dec_rt&1;
	  if(ch==cc)
	    fdl = fsl;
	  if(ch2==cc)
	    fdh = fsh;
	  set_fr64(dec_sa, fdh, fdl);
	}
	else{
	  except = Exc_RI;
	}
	break;
      case opMOVZ:
      case opMOVN:
	if(mips4_en){
	  get_gr(dec_rt, &fth, &ftl);
	  ch = ((fth|ftl)==0L);
	  cc = dec_fn&1;
	  if(ch!=cc)
	    set_fr64(dec_sa, fsh, fsl);
	}
	else{
	  except = Exc_RI;
	}
	break;
      case opCVT_S:/* CVT.S.PU */
        FPU(opNEG, fmtS, rm, fs, 0, fsh, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, 0, fp_neg(fdl), e, v, z, o, u, i, cmp, cc);
	break;
      case 044:/* CVT.PW.PS */
        FPU(opCVT_W, fmtS, rm, fs, 0, fsl, fth, ftl,
            &dd, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        FPU(opCVT_W, fmtS, rm, fs, 0, fsh, fth, ftl,
            &dd, &fdh, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
        check_set_fr64(dec_sa, fdh, fdl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
	break;
      case 050:/* CVT.S.PL */
        FPU(opNEG, fmtS, rm, fs, 0, fsl, fth, ftl,
            &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        check_set_fr64(dec_sa, 0, fp_neg(fdl), e, v, z, o, u, i, cmp, cc);
	break;
      case 054:/* PLL.PS */
#if 0
        FPU(opNEG, fmtS, rm, fs, 0, fsl, fth, ftl,
            &dd, &fdh, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        FPU(opNEG, fmtS, rm, fs, 0, ftl, fth, ftl,
            &dd, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
        check_set_fr64(dec_sa, fp_neg(fdh), fp_neg(fdl), e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
#else
/*	set_fr64(dec_sa,fsl,ftl);*/
	check_set_fr64(dec_sa, fsl, ftl, 0, 0, 0, 0, 0, 0, 0, 0);
#endif
	break;
      case 055:/* PLU.PS */
#if 0
        FPU(opNEG, fmtS, rm, fs, 0, fsl, fth, ftl,
            &dd, &fdh, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        FPU(opNEG, fmtS, rm, fs, 0, fth, fth, ftl,
            &dd, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
        check_set_fr64(dec_sa, fp_neg(fdh), fp_neg(fdl), e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
#else
/*	set_fr64(dec_sa,fsl,fth);*/
	check_set_fr64(dec_sa, fsl, fth, 0, 0, 0, 0, 0, 0, 0, 0);
#endif
	break;
      case 056:/* PUL.PS */
#if 0
        FPU(opNEG, fmtS, rm, fs, 0, fsh, fth, ftl,
            &dd, &fdh, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        FPU(opNEG, fmtS, rm, fs, 0, ftl, fth, ftl,
            &dd, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
        check_set_fr64(dec_sa, fp_neg(fdh), fp_neg(fdl), e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
#else
/*	set_fr64(dec_sa,fsh,ftl);*/
	check_set_fr64(dec_sa, fsh, ftl, 0, 0, 0, 0, 0, 0, 0, 0);
#endif
	break;
      case 057:/* PUU.PS */
#if 0
        FPU(opNEG, fmtS, rm, fs, 0, fsh, fth, ftl,
            &dd, &fdh, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	if(e){
	  v = z = o = u = i = 0;
	}
        FPU(opNEG, fmtS, rm, fs, 0, fth, fth, ftl,
            &dd, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	if(e2){
	  v2 = z2 = o2 = u2 = i2 = 0;
	}
        check_set_fr64(dec_sa, fp_neg(fdh), fp_neg(fdl), e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, cmp, cc);
#else
/*	set_fr64(dec_sa,fsh,fth);*/
	check_set_fr64(dec_sa, fsh, fth, 0, 0, 0, 0, 0, 0, 0, 0);
#endif
	break;
      case opADD:
      case opSUB:
      case opMUL:
      case opABS:
      case opNEG:
	PS_FPU(dec_fn, rm, fs, fsh, fsl, fth, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
	break;
      case 030:/* ADDR */
	PS_FPU(opADD, rm, fs, fsh, fth, fsl, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
	break;
      case 032:/* MULR */
	PS_FPU(opMUL, rm, fs, fsh, fth, fsl, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
	break;
      case 034:/* RECIP2 -(fs*ft-1)*/
	{
	  long rh,rl;
	  rh = 0x3f800000L;
	  rl = 0x3f800000L;
	  PS_FPUX(076/* NMSUB.PS */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
		  &e, &v, &z, &o, &u, &i);
	  check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, 0, 0);
	}
	break;
      case 035:/* RECIP1 */
	PS_FPU(opRECIP, rm, fs, fsh, fsl, fth, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
	break;
      case 036:/* RSQRT1 */
	PS_FPU(opRSQRT, rm, fs, fsh, fsl, fth, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
	break;
      case 037:/* RSQRT2 -(fs*ft-1)/2 */
	{
	  long rh,rl;
	  rh = 0x3f800000L;
	  rl = 0x3f800000L;
	  PS_FPUX(076/* NMSUB.PS */, rm, fs, fsh, fsl, fth, ftl, rh, rl, &fdh, &fdl,
		  &e, &v, &z, &o, &u, &i);
	  if(e){
	    v = z = o = u = i = 0;
	  }
	  else{
	    if(v){
	      if(fdl==0x7fbfffff)
		fdl = 0x7fffffff;
	      if(fdh==0x7fbfffff)
		fdh = 0x7fffffff;
	    }
	    fsh = fdh;
	    fsl = fdl;
	    fth = 0x40000000L;
	    ftl = 0x40000000L;
	    PS_FPU(opDIV, rm, fs, fsh, fsl, fth, ftl,
		   &fdh, &fdl, &e2, &v2, &z2, &o2, &u2, &i2, &cmp, &cc);
	    if(e2){
	      v2 = z2 = o2 = u2 = i2 = 0;
	    }
	    check_set_fr64(dec_sa, fdh, fdl, e|e2, v|v2, z|z2, o|o2, u|u2, i|i2, 0, 0);
	  }
	}
	break;
      case opCMP_F:
      case opCMP_UN:
      case opCMP_EQ:
      case opCMP_UEQ:
      case opCMP_OLT:
      case opCMP_ULT:
      case opCMP_OLE:
      case opCMP_ULE:
      case opCMP_SF:
      case opCMP_NGLE:
      case opCMP_SEQ:
      case opCMP_NGL:
      case opCMP_LT:
      case opCMP_NGE:
      case opCMP_LE:
      case opCMP_NGT:
	if(m3d_en&&(dec_sa&1)){/* CABS */
	  fsh &= 0x7fffffffL;
	  fth &= 0x7fffffffL;
	  fsl &= 0x7fffffffL;
	  ftl &= 0x7fffffffL;
	}
	PS_FPU(dec_fn, rm, fs, fsh, fsl, fth, ftl,
	       &fdh, &fdl, &e, &v, &z, &o, &u, &i, &cmp, &cc);
#if 0
	check_set_fr64(dec_sa, fdh, fdl, e, v, z, o, u, i, cmp, cc);
#else
	check_set_fr64(dec_sa, fdh, fdl, e, v, 0, 0, 0, 0, cmp, cc);
#endif
	break;
      default:
	except = Exc_RI;
      }
    }
    else{
      check_set_fr32(0, 0L, 1, 0, 0, 0, 0, 0, 0, 0);
    }
    break;
  default:
    check_set_fr32(0, 0L, 1, 0, 0, 0, 0, 0, 0, 0);
  }
}

static void exec_cop1x(void)
{
  ulong base_hi, base_lo;
  ulong data_hi, data_lo;
  ulong sh,sl,th,tl,rh,rl,dh,dl;
  int   fs,rm;
  int   e,v,z,o,u,i;
  int   cy;

  if(!mips4_en){
    except = Exc_RI;
    return;
  }
  if(!cp1_en){
#if 0
    except = Exc_CpU;
#else
    except = Exc_RI;
#endif
    return;
  }
  fs = ((FCR31&((ulong)(1<<24)))!=0L);
  rm = FCR31&3;
  switch(dec_fn){
  case 000:/* LWXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    if(base_lo&0x3){
      except = Exc_AdEL;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
        set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
        dst_lo = data_lo;
        if(data_lo&0x80000000L){
          dst_hi = 0xffffffffL;
        }
        else{
          dst_hi = 0L;
        }
	if(fr_bit)
	  set_fr64(dec_sa, dst_hi, dst_lo);
	else
	  set_fr32(dec_sa, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 001:/* LDXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    if(base_lo&7L){
      except = Exc_AdEL;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
        set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
        data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);
        dst_lo = data_lo;
        dst_hi = data_hi;
        set_fr64(dec_sa, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 005:/* LUXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);
      dst_lo = data_lo;
      dst_hi = data_hi;
      set_fr64(dec_sa, dst_hi, dst_lo);
      LLbit = 0;
    }
    break;
  case 010:/* SWXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    if(base_lo&0x3){
      except = Exc_AdES;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
        set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
        dst_lo = get_fr32(dec_rd);
        put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 011:/* SDXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    if(base_lo&7L){
      except = Exc_AdES;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
        set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
        data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
        get_fr64(dec_rd, &dst_hi, &dst_lo);
        put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
        put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
	LLbit = 0;
      }
    }
    break;
  case 015:/* SUXC1 */
    IADD64(src_hi, src_lo, tar_hi, tar_lo, &base_hi, &base_lo, &cy);
    data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
      get_fr64(dec_rd, &dst_hi, &dst_lo);
      put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
      put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
      LLbit = 0;
    }
    break;
  case 017:/* PREFX */
    break;
  case 036:/* ALNV.PS */
    get_fr64(dec_rd, &sh, &sl);
    get_fr64(dec_rt, &th, &tl);
    get_gr(dec_rs, &rh, &rl);
    rl &= 7;
#if 1 /* LittleEndian */
    if(rl==0){
      dl = sl;
      dh = sh;
    }
    else if(rl<=4){
      ulong t;
      int cnt;
      cnt = rl<<3;
      RSHIFTL(sh,sl,cnt,&dh,&dl);/* dl is fixed. */
      LSHIFT (th,tl,64-cnt,&t,&rh,&rl);
      dh |= rh;
    }
    else{
      ulong t;
      int cnt;
      cnt = (rl-4)<<3;
      RSHIFTL(tl,sh,cnt,&dh,&dl);/* dl is fixed. */
      LSHIFT (th,th,64-cnt,&t,&rh,&rl);
      dh |= rh;
    }
#else /* BigEndian */
    if(rl==0){
      dl = sl;
      dh = sh;
    }
    else if(rl<=4){
      ulong t;
      int cnt;
      cnt = rl<<3;
      LSHIFT (sh,sl,cnt,&t,&dh,&dl);/* dh is fixed. */
      RSHIFTL(th,tl,64-cnt,&rh,&rl);
      dl |= rl;
    }
    else{
      ulong t;
      int cnt;
      cnt = (rl-4)<<3;
      LSHIFT (sl,th,cnt,&t,&dh,&dl);/* dl is fixed. */
      RSHIFTL(tl,tl,64-cnt,&rh,&rl);
      dl |= rl;
    }
#endif
    check_set_fr64(dec_sa, dh, dl, 0, 0, 0, 0, 0, 0, 0, 0);
    break;
  case 040:/* MADD.S */
  case 050:/* MSUB.S */
  case 060:/* NMADD.S */
  case 070:/* NMSUB.S */
    sl = get_fr32(dec_rd);
    tl = get_fr32(dec_rt);
    rl = get_fr32(dec_rs);
    FPUX(dec_fn, rm, fs, 0L, sl, 0L, tl, 0L, rl, &dh, &dl,
         &e, &v, &z, &o, &u, &i);
    if(e){
      v = z = o = u = i = 0;
    }
    check_set_fr64(dec_sa, 0, dl, e, v, z, o, u, i, 0, 0);
    break;
  case 041:/* MADD.D */
  case 051:/* MSUB.D */
  case 061:/* NMADD.D */
  case 071:/* NMSUB.D */
    get_fr64(dec_rd, &sh, &sl);
    get_fr64(dec_rt, &th, &tl);
    get_fr64(dec_rs, &rh, &rl);
#if 0
    printf("(s, t, r)=(%08x %08x, %08x %08x, %08x %08x)\n",sh,sl,th,tl,rh,rl);
#endif
    FPUX(dec_fn, rm, fs, sh, sl, th, tl, rh, rl, &dh, &dl,
         &e, &v, &z, &o, &u, &i);
#if 0
    printf("d=%08x %08x\n",dh,dl);
#endif
    if(e){
      v = z = o = u = i = 0;
    }
    check_set_fr64(dec_sa, dh, dl, e, v, z, o, u, i, 0, 0);
    break;
  case 046:/* MADD.PS */
  case 056:/* MSUB.PS */
  case 066:/* NMADD.PS */
  case 076:/* NMSUB.PS */
    get_fr64(dec_rd, &sh, &sl);
    get_fr64(dec_rt, &th, &tl);
    get_fr64(dec_rs, &rh, &rl);
    PS_FPUX(dec_fn, rm, fs, sh, sl, th, tl, rh, rl, &dh, &dl,
         &e, &v, &z, &o, &u, &i);
    check_set_fr64(dec_sa, dh, dl, e, v, z, o, u, i, 0, 0);
    break;
  default:
    check_set_fr32(0, 0L, 1, 0, 0, 0, 0, 0, 0, 0);
  }
}

static void exec_cop0(void)
{
  ulong base_hi, base_lo;
  ulong immd_hi, immd_lo;
  ulong v;
  int   cy,ch,erl;

  if(!cp0_en){
    except = Exc_CpU;
    return;
  }
  switch(dec_rs){
  case 000:/* MF */
    dec_rd |= ((dec_fn&7)<<5);
    if(addr64){
      dmfc0(dec_rd, &dst_hi, &dst_lo);
    }
    else{
      dst_lo = mfc0(dec_rd);
      dst_hi = sign_extend(dst_lo);
    }
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 001:/* DMF */
   dec_rd |= ((dec_fn&7)<<5);
   if(inst64_en){
      dmfc0(dec_rd, &dst_hi, &dst_lo);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 002:/* CF */
    except = Exc_RI;
    break;
  case 004:/* MT */
    dec_rd |= ((dec_fn&7)<<5);
    if(addr64)
      dmtc0(dec_rd, tar_hi, tar_lo);
    else
      mtc0(dec_rd, tar_lo);
    break;
  case 005:/* DMT */
    dec_rd |= ((dec_fn&7)<<5);
    if(inst64_en){
      dmtc0(dec_rd, tar_hi, tar_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 006:/* CT */
    except = Exc_RI;
    break;
  case 010:/* BC */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ch = ((mfc0(C0_SR)&((ulong)(1<<18)))!=0L);
    br_inst = 4;
    switch(dec_rt){
    case 000:/* BCF */
      br_cond    = (ch==0);
      br_likely  = 0;
      break;
    case 001:/* BCT */
      br_cond    = (ch==1);
      br_likely  = 0;
      break;
    case 002:/* BCFL */
      br_cond    = (ch==0);
      br_likely  = 1;
      break;
    case 003:/* BCTL */
      br_cond    = (ch==1);
      br_likely  = 1;
      break;
    default:
      except = Exc_RI;
    }
    break;
  case 013:/* EI/DI */
    if(ver2_en && (dec_rd==0xc)){
      dst_lo = mfc0(12);/* C0_SR */
      dst_hi = sign_extend(dst_lo);
      set_gr(dec_rt, dst_hi, dst_lo);
      dst_lo |= 1;
      if((dec_fn&0x20)==0)
	dst_lo ^= 1;
      mtc0(12, dst_lo);
    }
    break;
  case 020:/* CO */
  case 021:
  case 022:
  case 023:
  case 024:
  case 025:
  case 026:
  case 027:
  case 030:
  case 031:
  case 032:
  case 033:
  case 034:
  case 035:
  case 036:
  case 037:
    switch(dec_fn){
    case 001:/* TLBR */
      tlbr();
      break;
    case 002:/* TLBWI */
      tlbwi();
      break;
    case 006:/* TLBWR */
      tlbwr();
      break;
    case 010:/* TLBP */
      tlbp();
      break;
    case 030:/* ERET */
      v = mfc0(C0_SR);
      erl = ((v&4L)!=0L);
      if(erl){
	dmfc0(C0_ErrorEPC, &target_hi, &target_lo);
	v &= 0xfffffffbL;
      }
      else{
	dmfc0(C0_EPC, &target_hi, &target_lo);
	v &= 0xfffffffdL;
      }
      no_ds_br = 1;
      LLbit = 0;
      mtc0(C0_SR, v);
      break;
    }
    break;
  default:
    except = Exc_RI;
  }
}

static void exec_special(void)
{
  ulong udummy;
  int   idummy;
  int    ssign;
  int    tsign;
  int    dsign;
  int    lt, eq;
  int    ch, cc;

  switch(dec_fn){
  case 000:/* SLL */
    LSHIFT(tar_hi, tar_lo, dec_sa, &udummy, &dst_hi, &dst_lo);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 001:/* MOVCI */
    if(mips4_en){
      if(cp1_en){
        cc = (dec_rt>>2)&7;
        if(cc==0)
          ch = ((FCR31&((ulong)(1<<23)))!=0L);
        else
          ch = ((FCR31&((ulong)(1<<(cc+24))))!=0L);
        cc = dec_rt&1;
        if(ch==cc)
          set_gr(dec_rd, src_hi, src_lo);
      }
      else{
	except = Exc_CpU;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 002:/* SRL */
    RSHIFTL(0L,  tar_lo, dec_sa, &dst_hi, &dst_lo);
    if(dec_rs&1){/* ROR */
      ulong dst_hi2,dst_lo2;
      LSHIFT(tar_hi, tar_lo, 32-dec_sa, &udummy, &dst_hi2, &dst_lo2);
      dst_lo = dst_lo | dst_lo2;
    }
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 003:/* SRA */
    RSHIFTA(sign_extend(tar_lo),  tar_lo, dec_sa, &dst_hi, &dst_lo);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 004:/* SLLV */
    idummy = src_lo & 0x1f;
    LSHIFT(tar_hi, tar_lo, idummy, &udummy, &dst_hi, &dst_lo);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 006:/* SRLV */
    idummy = src_lo & 0x1f;
    RSHIFTL(0L,  tar_lo, idummy, &dst_hi, &dst_lo);
    if(dec_sa&1){/* RORV */
      ulong dst_hi2,dst_lo2;
      LSHIFT(tar_hi, tar_lo, 32-idummy, &udummy, &dst_hi2, &dst_lo2);
      dst_lo = dst_lo | dst_lo2;
    }
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 007:/* SRAV */
    idummy = src_lo & 0x1f;
    RSHIFTA(sign_extend(tar_lo),  tar_lo, idummy, &dst_hi, &dst_lo);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
   break;
  case 010:/* JR */
    target_hi = src_hi;
    target_lo = src_lo;
    br_likely  = 0;
    br_cond    = 1;
    br_inst = 4;
    break;
  case 011:/* JALR */
    target_hi = src_hi;
    target_lo = src_lo;
    IADD64(PCH, PCL, 0L, 8L, &dst_hi, &dst_lo, &idummy);
    set_gr(dec_rd, dst_hi, dst_lo);
    br_likely  = 0;
    br_cond    = 1;
    br_inst = 4;
    break;
  case 012:/* MOVZ */
    if(mips4_en){
      if((tar_hi|tar_lo)==0L){
	set_gr(dec_rd, src_hi, src_lo);
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 013:/* MOVN */
    if(mips4_en){
      if(tar_hi|tar_lo){
	set_gr(dec_rd, src_hi, src_lo);
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 014:/* SYSCALL */
    except = Exc_Sys;
    break;
  case 015:/* BREAK */
    except = Exc_Bp;
    break;
  case 017:/* SYNC */
    break;
  case 020:/* MFHI */
    set_gr(dec_rd, HIHI, HILO);
    break;
  case 021:/* MTHI */
    set_hi(src_hi, src_lo);
    break;
  case 022:/* MFLO */
    set_gr(dec_rd, LOHI, LOLO);
    break;
  case 023:/* MTLO */
    set_lo(src_hi, src_lo);
    break;
  case 024:/* DSLLV */
    if(inst64_en){
      idummy = src_lo&0x3f;
      LSHIFT(tar_hi, tar_lo, idummy, &udummy, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 026:/* DSRLV */
    if(inst64_en){
      idummy = src_lo&0x3f;
      RSHIFTL(tar_hi, tar_lo, idummy, &dst_hi, &dst_lo);
      if(dec_sa&1){/* DRORV */
	ulong dst_hi2,dst_lo2;
	LSHIFT(tar_hi, tar_lo, 64-idummy, &udummy, &dst_hi2, &dst_lo2);
	dst_lo |= dst_lo2;
	dst_hi |= dst_hi2;
      }
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 027:/* DSRAV */
    if(inst64_en){
      idummy = src_lo&0x3f;
      RSHIFTA(tar_hi, tar_lo, idummy, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 030:/* MULT */
    if(dec_sa){
      exec_macc5432((dec_sa<<6)|030, src_lo, tar_lo, &dst_lo);
      dst_hi = sign_extend(dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
	     &HIHI, &HILO, &LOHI, &LOLO);
      HILO = LOHI;
      LOHI = sign_extend(LOLO);
      HIHI = sign_extend(HILO);
    }
    break;
  case 031:/* MULTU */
    if(dec_sa){
      exec_macc5432((dec_sa<<6)|031, src_lo, tar_lo, &dst_lo);
      dst_hi = sign_extend(dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      UMUL32(src_lo, tar_lo, &HILO, &LOLO);
      HIHI = sign_extend(HILO);
      LOHI = sign_extend(LOLO);
    }
    break;
  case 032:/* DIV */
    ssign = ((src_lo&0x80000000L)!=0);
    tsign = ((tar_lo&0x80000000L)!=0);
    dsign = ssign ^ tsign;
    if(ssign){
      src_lo = ~src_lo+1L;
    }
    if(tsign){
      tar_lo = ~tar_lo+1L;
    }
    UDIV32(src_lo, tar_lo,&LOLO, &HILO);
    if(dsign){
      LOLO = ~LOLO+1L;
    }
    if(ssign){
      HILO = ~HILO+1L;
    }
    LOHI = sign_extend(LOLO);
    HIHI = sign_extend(HILO);
    break;
  case 033:/* DIVU */
    UDIV32(src_lo, tar_lo,&LOLO, &HILO);
    LOHI = sign_extend(LOLO);
    HIHI = sign_extend(HILO);
    break;
  case 034:/* DMULT */
    if(inst64_en){
      ssign = ((src_hi&0x80000000L)!=0);
      tsign = ((tar_hi&0x80000000L)!=0);
      dsign = ssign ^ tsign;
      if(ssign){
        ISUB64(0L, 0L, src_hi, src_lo, &dst_hi, &dst_lo, &idummy);
        src_hi = dst_hi;
        src_lo = dst_lo;
      }
      if(tsign){
        ISUB64(0L, 0L, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
        tar_hi = dst_hi;
        tar_lo = dst_lo;
      }
      UMUL64(src_hi, src_lo, tar_hi, tar_lo, &HIHI, &HILO, &LOHI, &LOLO);
      if(dsign){
        NEG128(HIHI, HILO, LOHI, LOLO, &HIHI, &HILO, &LOHI, &LOLO);
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 035:/* DMULTU */
    if(inst64_en){
      UMUL64(src_hi, src_lo, tar_hi, tar_lo, &HIHI, &HILO, &LOHI, &LOLO);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 036:/* DDIV */
    if(inst64_en){
      ssign = ((src_hi&0x80000000L)!=0);
      tsign = ((tar_hi&0x80000000L)!=0);
      dsign = ssign ^ tsign;
      if(ssign){
        ISUB64(0L, 0L, src_hi, src_lo, &dst_hi, &dst_lo, &idummy);
        src_hi = dst_hi;
        src_lo = dst_lo;
      }
      if(tsign){
        ISUB64(0L, 0L, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
        tar_hi = dst_hi;
        tar_lo = dst_lo;
      }
      UDIV64(src_hi, src_lo, tar_hi, tar_lo, &LOHI, &LOLO, &HIHI, &HILO);
      if(dsign){
        ISUB64(0L, 0L, LOHI, LOLO, &LOHI, &LOLO, &idummy);
      }
      if(ssign){
        ISUB64(0L, 0L, HIHI, HILO, &HIHI, &HILO, &idummy);
      }
    }
    else{
      except = Exc_RI;
    }
     break;
  case 037:/* DDIVU */
    if(inst64_en){
      UDIV64(src_hi, src_lo, tar_hi, tar_lo, &LOHI, &LOLO, &HIHI, &HILO);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 040:/* ADD */
    IADD64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &dst_hi, &dst_lo, &idummy);
    dsign = ((dst_lo&0x80000000L)!=0);
    idummy = dst_hi&1;
    if(dsign==idummy){
      set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    }
    else{
      except = Exc_Ov;
    }
    break;
  case 041:/* ADDU */
    IADD64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &dst_hi, &dst_lo, &idummy);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 042:/* SUB */
    ISUB64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &dst_hi, &dst_lo, &idummy);
    dsign = ((dst_lo&0x80000000L)!=0);
    idummy = dst_hi&1;
    if(dsign==idummy){
      set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    }
    else{
      except = Exc_Ov;
    }
    break;
  case 043:/* SUBU */
    ISUB64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &dst_hi, &dst_lo, &idummy);
    set_gr(dec_rd, sign_extend(dst_lo), dst_lo);
    break;
  case 044:/* AND */
    dst_hi = src_hi & tar_hi;
    dst_lo = src_lo & tar_lo;
    set_gr(dec_rd, dst_hi, dst_lo);
    break;
  case 045:/* OR */
    dst_hi = src_hi | tar_hi;
    dst_lo = src_lo | tar_lo;
    set_gr(dec_rd, dst_hi, dst_lo);
    break;
  case 046:/* XOR */
    dst_hi = src_hi ^ tar_hi;
    dst_lo = src_lo ^ tar_lo;
    set_gr(dec_rd, dst_hi, dst_lo);
    break;
  case 047:/* NOR */
    dst_hi = src_hi | tar_hi;
    dst_lo = src_lo | tar_lo;
    set_gr(dec_rd, ~dst_hi, ~dst_lo);
   break;
  case 050:/* MACC (R4100) */
    if(ty_cpu==R4100){
      int sat, hi, us;
      sat = ((dec_imm&0x400)!=0);
      hi  = ((dec_imm&0x200)!=0);
      us  = ((dec_imm&0x040)!=0);
      exec_macc(sat, hi, us);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 051:/* DMACC (R4100) */
    if((ty_cpu==R4100)&&inst64_en){
      int sat, hi, us;
      sat = ((dec_imm&0x400)!=0);
      hi  = ((dec_imm&0x200)!=0);
      us  = ((dec_imm&0x040)!=0);
      exec_dmacc(sat, hi, us);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 052:/* SLT */
    CMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if(lt&&(!eq)){
      dst_lo = 1L;
    }
    else{
      dst_lo = 0L;
    }
    set_gr(dec_rd, 0L, dst_lo);
    break;
  case 053:/* SLTU */
    UCMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if(lt&&(!eq)){
      dst_lo = 1L;
    }
    else{
      dst_lo = 0L;
    }
    set_gr(dec_rd, 0L, dst_lo);
    break;
  case 054:/* DADD */
    if(inst64_en){
      ssign = ((src_hi&0x80000000L)!=0L);
      tsign = ((tar_hi&0x80000000L)!=0L);
      IADD64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
      dsign = ((dst_hi&0x80000000L)!=0);
      idummy = (idummy+ssign+tsign)&1;
      if(dsign==idummy){
        set_gr(dec_rd, dst_hi, dst_lo);
      }
      else{
        except = Exc_Ov;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 055:/* DADDU */
    if(inst64_en){
      IADD64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 056:/* DSUB */
    if(inst64_en){
      ssign = ((src_hi&0x80000000L)!=0L);
      tsign = ((tar_hi&0x80000000L)==0L);
      ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
      dsign = ((dst_hi&0x80000000L)!=0);
      idummy = (idummy+ssign+tsign)&1;
      if(dsign==idummy){
        set_gr(dec_rd, dst_hi, dst_lo);
      }
      else{
        except = Exc_Ov;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 057:/* DSUBU */
    if(inst64_en){
      ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &idummy);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 060:/* TGE */
    CMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if((!lt)||eq){
      except = Exc_Tr;
    }
    break;
  case 061:/* TGEU */
    UCMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if((!lt)||eq){
      except = Exc_Tr;
    }
   break;
  case 062:/* TLT */
    CMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if((lt)&&(!eq)){
      except = Exc_Tr;
    }
    break;
  case 063:/* TLTU */
    UCMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if((lt)&&(!eq)){
      except = Exc_Tr;
    }
    break;
  case 064:/* TEQ */
    CMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if(eq){
      except = Exc_Tr;
    }
    break;
  case 066:/* TNE */
    CMP64(src_hi, src_lo, tar_hi, tar_lo, &lt, &eq);
    if(!eq){
      except = Exc_Tr;
    }
    break;
  case 070:/* DSLL */
    if(inst64_en){
      LSHIFT(tar_hi, tar_lo, dec_sa, &udummy, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 072:/* DSRL */
    if(inst64_en){
      RSHIFTL(tar_hi, tar_lo, dec_sa, &dst_hi, &dst_lo);
      if(dec_rs&1){/* DROR */
	ulong dst_hi2,dst_lo2;
	LSHIFT(tar_hi, tar_lo, 64-dec_sa, &udummy, &dst_hi2, &dst_lo2);
	dst_lo |= dst_lo2;
	dst_hi |= dst_hi2;
      }
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 073:/* SDRA */
    if(inst64_en){
      RSHIFTA(tar_hi, tar_lo, dec_sa, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 074:/* DSLL32 */
    if(inst64_en){
      idummy = dec_sa+32;
      LSHIFT(tar_hi, tar_lo, idummy, &udummy, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 076:/* DSRL32 */
    if(inst64_en){
      idummy = dec_sa+32;
      RSHIFTL(tar_hi, tar_lo, idummy, &dst_hi, &dst_lo);
      if(dec_rs&1){
	ulong dst_hi2,dst_lo2;
	LSHIFT(tar_hi, tar_lo, 64-idummy, &udummy, &dst_hi2, &dst_lo2);
	dst_lo |= dst_lo2;
	dst_hi |= dst_hi2;
      }
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 077:/* DSRA32 */
    if(inst64_en){
      idummy = dec_sa+32;
      RSHIFTA(tar_hi, tar_lo, idummy, &dst_hi, &dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  default:/* Reserved Instruction */
    except = Exc_RI;
  }
}

static void exec_regimm(void)
{
  ulong imm_hi, imm_lo;
  ulong link_hi, link_lo;
  ulong targ_hi_tmp,targ_lo_tmp;
  ulong dummy1, dummy2, dummy3;
  int   cy;
  int   lt, eq;

  if(dec_imm&0x8000){
    imm_hi = 0xffffffffL;
    imm_lo = 0xffff0000L|dec_imm;
  }
  else{
    imm_hi = 0L;
    imm_lo = dec_imm;
  }
  LSHIFT(imm_hi, imm_lo, 2, &dummy1, &dummy2, &dummy3);
  IADD64(PCH, PCL, dummy2, dummy3, &dummy2, &dummy3, &cy);
  IADD64(dummy2, dummy3, 0L, 4L, &targ_hi_tmp, &targ_lo_tmp, &cy);
  IADD64(PCH, PCL, 0L, 8L, &link_hi, &link_lo, &cy);

  except = Exc_NONE;
  switch(dec_rt){
  case 000:/* BLTZ */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 0;
    br_cond    = ((src_hi&0x80000000L)!=0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 001:/* BGEZ */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 0;
    br_cond    = ((src_hi&0x80000000L)==0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 002:/* BLTZL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 1;
    br_cond    = ((src_hi&0x80000000L)!=0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 003:/* BGEZL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 1;
    br_cond    = ((src_hi&0x80000000L)==0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 010:/* TGEI */
    CMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if((!lt)||eq){
      except = Exc_Tr;
    }
    break;
  case 011:/* TGEIU */
    UCMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if((!lt)||eq){
      except = Exc_Tr;
    }
    break;
  case 012:/* TLTI */
    CMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if(lt&&(!eq)){
      except = Exc_Tr;
    }
    break;
  case 013:/* TLTIU */
    UCMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if(lt&&(!eq)){
      except = Exc_Tr;
    }
    break;
  case 014:/* TEQI */
    CMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if(eq){
      except = Exc_Tr;
    }
    break;
  case 016:/* TNEI */
    CMP64(src_hi, src_lo, imm_hi, imm_lo, &lt, &eq);
    if(!eq){
      except = Exc_Tr;
    }
    break;
  case 020:/* BLTZAL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 0;
    br_cond    = ((src_hi&0x80000000L)!=0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    set_gr(31, link_hi, link_lo);
    break;
  case 021:/* BGEZAL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 0;
    br_cond    = ((src_hi&0x80000000L)==0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    set_gr(31, link_hi, link_lo);
    break;
  case 022:/* BLTZALL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 1;
    br_cond    = ((src_hi&0x80000000L)!=0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    set_gr(31, link_hi, link_lo);
    break;
  case 023:/* BGEZALL */
    target_hi = targ_hi_tmp;
    target_lo = targ_lo_tmp;
    br_likely  = 1;
    br_cond    = ((src_hi&0x80000000L)==0L);
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    set_gr(31, link_hi, link_lo);
    break;
  case 037:/* SYNCI */
    if(ver2_en){
      get_gr(dec_rs, &dummy1, &dummy2);
      IADD64(imm_hi, imm_lo, dummy1, dummy2, &dummy1, &dummy2, &cy);
      Hit_WB_I(dummy1, dummy2, &except);
      if(except == Exc_NONE)
        Hit_Inv_I(dummy1, dummy2, &except);
    }
    else{
      except = Exc_RI;
    }
    break;
  default:/* Reserved Instruction */
    except = Exc_RI;
  }
}

static void do_restore(int mode)
{
  int ra,s0s1,sreg,areg,size;
  ulong extend,insn;
  size =((dec_rd<<5)|dec_sa);
  s0s1 = dec_rt&3;
  ra   = (dec_rt>>2)&1;
  areg = ((dec_rs&3)<<2)|((dec_rt>>3)&3);
  sreg = (dec_rs>>2)&7;
  insn = (0<<7)|(ra<<6)|(s0s1<<4)|(size&0xf);
  extend = (sreg<<8)|(areg)|(size&0xf0)|0xf000;
  do_save_restore(extend, insn, mode);
}

static void do_save(int mode)
{
  int ra,s0s1,sreg,areg,size;
  ulong extend,insn;
  size =((dec_rd<<5)|dec_sa);
  s0s1 = dec_rt&3;
  ra   = (dec_rt>>2)&1;
  areg = ((dec_rs&3)<<2)|((dec_rt>>3)&3);
  sreg = (dec_rs>>2)&7;
  insn = (1<<7)|(ra<<6)|(s0s1<<4)|(size&0xf);
  extend = (sreg<<8)|(areg)|(size&0xf0)|0xf000;
  do_save_restore(extend, insn, mode);
}

static void exec_special2(void)
{
  ulong m3, m2, m1, m0;
  int   cy;

  if(!mips32_en){
    except = Exc_RI;
    return;
  }

  switch(dec_fn){
  case 000:/* MADD */
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &m3, &m2, &m1, &m0);
    IADD64(m1, m0, HILO, LOLO, &dst_hi, &dst_lo, &cy);
    LOLO = dst_lo;
    LOHI = sign_extend(dst_lo);
    HILO = dst_hi;
    HIHI = sign_extend(dst_hi);
    break;
  case 001:/* MADDU */
    UMUL64(0L, src_lo, 0L, tar_lo,
           &m3, &m2, &m1, &m0);
    IADD64(m1, m0, HILO, LOLO, &dst_hi, &dst_lo, &cy);
    LOLO = dst_lo;
    LOHI = sign_extend(dst_lo);
    HILO = dst_hi;
    HIHI = sign_extend(dst_hi);
    break;
  case 002:/* MUL */
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &m3, &m2, &m1, &m0);
    set_gr(dec_rd, sign_extend(m0), m0);
    LOLO = m0;
    LOHI = sign_extend(m0);
    HILO = m1;
    HIHI = sign_extend(m1);
    break;
  case 004:/* MSUB */
    UMUL64(sign_extend(src_lo), src_lo, sign_extend(tar_lo), tar_lo,
           &m3, &m2, &m1, &m0);
    ISUB64(HILO, LOLO, m1, m0, &dst_hi, &dst_lo, &cy);
    LOLO = dst_lo;
    LOHI = sign_extend(dst_lo);
    HILO = dst_hi;
    HIHI = sign_extend(dst_hi);
    break;
  case 005:/* MSUBU */
    UMUL64(0L, src_lo, 0L, tar_lo,
           &m3, &m2, &m1, &m0);
    ISUB64(HILO, LOLO, m1, m0, &dst_hi, &dst_lo, &cy);
    LOLO = dst_lo;
    LOHI = sign_extend(dst_lo);
    HILO = dst_hi;
    HIHI = sign_extend(dst_hi);
    break;
  case 040:/* CLZ */
    dst_lo = CLZ32(src_lo);
#if 0
    set_gr(dec_rt, 0L, dst_lo);
#else
    set_gr(dec_rd, 0L, dst_lo);
#endif
    break;
  case 041:/* CLO */
    dst_lo = CLO32(src_lo);
#if 0
    set_gr(dec_rt, 0L, dst_lo);
#else
    set_gr(dec_rd, 0L, dst_lo);
#endif
    break;
  case 044:/* DCLZ */
    if(inst64_en){
      dst_lo = CLZ64(src_hi, src_lo);
#if 0
      set_gr(dec_rt, 0L, dst_lo);
#else
      set_gr(dec_rd, 0L, dst_lo);
#endif
    }
    else{
      except = Exc_RI;
    }
    break;
  case 045:/* DCLO */
    if(inst64_en){
      dst_lo = CLO64(src_hi, src_lo);
#if 0
      set_gr(dec_rt, 0L, dst_lo);
#else
      set_gr(dec_rd, 0L, dst_lo);
#endif
    }
    else{
      except = Exc_RI;
    }
    break;
  default:/* Reserved Instruction */
    except = Exc_RI;
  }
}

static void exec_cache(int op, ulong hi, ulong lo)
{
  switch(op){
  case /* Index_Inv_I           */ 0x0:
    Index_Inv_I(hi, lo, &except);
    break;
  case /* Index_Writeback_Inv_D */ 0x1:
    Index_WB_Inv_D(hi, lo, &except);
    break;
  case /* Index_Load_Tag_I      */ 0x4:
    Index_Load_Tag_I(hi, lo, &except);
    break;
  case /* Index_Load_Tag_D      */ 0x5:
    Index_Load_Tag_D(hi, lo, &except);
    break;
  case /* Index_Store_Tag_I     */ 0x8:
    Index_Store_Tag_I(hi, lo, &except);
    break;
  case /* Index_Store_Tag_D     */ 0x9:
    Index_Store_Tag_D(hi, lo, &except);
    break;
  case /* Create_Dirty_D	*/ 0xd:
    Create_Dirty_D(hi, lo, &except);
    break;
  case /* Hit_Inv_I             */ 0x10:
    Hit_Inv_I(hi, lo, &except);
    break;
  case /* Hit_Inv_D             */ 0x11:
    Hit_Inv_D(hi, lo, &except);
    break;
  case /* Hit_Writeback_Inv_D   */ 0x15:
    Hit_WB_Inv_D(hi, lo, &except);
  case /* Fill_I	        */ 0x14:
    Fill_I(hi, lo, &except);
    break;
  case /* Hit_Writeback_D	*/ 0x19:
    Hit_WB_D(hi, lo, &except);
    break;
  case /* Hit_Writeback_I	*/ 0x18:
    Hit_WB_I(hi, lo, &except);
    break;
  }
}

static void exec_special3(void)
{
  ulong mask;
  ulong mask2;
  ulong mask3;
  ulong mask4;
  ulong mask5;
  int  cy;
  switch(dec_fn){
  case 000:/* EXT */
    if(dec_rd==31)
      mask = 0xffffffff;
    else
      mask = (1<<(dec_rd+1))-1;
    dst_lo = ((ulong)src_lo >> dec_sa) & mask;
    set_gr(dec_rt, 0, dst_lo);
    break;
  case 001:/* DEXTM */
    if(inst64_en){
      if(dec_rd==31){
        mask  = 0xffffffff;
        mask2 = 0xffffffff;
      }
      else{
        LSHIFT(0, 1, dec_rd+33, &mask3, &mask2, &mask);
        ISUB64(mask2, mask, 0, 1, &mask2, &mask, &cy);
      }
      RSHIFTL(src_hi, src_lo, dec_sa, &dst_hi, &dst_lo);
      dst_lo &= mask;
      dst_hi &= mask2;
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 002:/* DEXTU */
    if(inst64_en){
      if(dec_rd==31)
        mask = 0xffffffff;
      else
        mask = (1<<(dec_rd+1))-1;
      RSHIFTL(src_hi, src_lo, dec_sa+32, &dst_hi, &dst_lo);
      dst_lo &= mask;
      set_gr(dec_rt, 0, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 003:/* DEXT */
    if(inst64_en){
      if(dec_rd==31)
        mask = 0xffffffff;
      else
        mask = (1<<(dec_rd+1))-1;
      RSHIFTL(src_hi, src_lo, dec_sa, &dst_hi, &dst_lo);
      dst_lo &= mask;
      set_gr(dec_rt, 0, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 004:/* INS */
    if(dec_rd==31)
      mask = 0xffffffff;
    else
      mask = (1<<(dec_rd+1))-1;
    mask2 = (1<<dec_sa)-1;
    mask  = (~mask)|mask2;
    mask2 = ~mask;
    dst_lo = (tar_lo & mask)|((src_lo<<dec_sa)&mask2);
    dst_hi = tar_hi;
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 005:/* DINSM */
    if(inst64_en){
      if(dec_rd==31){
        mask  = 0xffffffff;
        mask2 = 0xffffffff;
      }
      else{
        LSHIFT(0, 1, dec_rd+33, &mask5, &mask2, &mask);
        ISUB64(mask2, mask, 0, 1, &mask2, &mask, &cy);
      }
      LSHIFT(0, 1, dec_sa, &mask5, &mask4, &mask3);
      ISUB64(mask4, mask3, 0, 1, &mask4, &mask3, &cy);
      mask  = (~mask) |mask3;
      mask2 = (~mask2)|mask4;
      mask3 = ~mask;
      mask4 = ~mask2;
      LSHIFT(src_hi, src_lo, dec_sa, &mask5, &dst_hi, &dst_lo);
      dst_lo = (dst_lo & mask3) | (tar_lo & mask);
      dst_hi = (dst_hi & mask4) | (tar_hi & mask2);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 006:/* DINSU */
    if(inst64_en){
      if(dec_rd==31){
        mask  = 0xffffffff;
        mask2 = 0xffffffff;
      }
      else{
        LSHIFT(0, 1, dec_rd+33, &mask5, &mask2, &mask);
        ISUB64(mask2, mask, 0, 1, &mask2, &mask, &cy);
      }
      LSHIFT(0, 1, dec_sa+32, &mask5, &mask4, &mask3);
      ISUB64(mask4, mask3, 0, 1, &mask4, &mask3, &cy);
      mask  = (~mask) |mask3;
      mask2 = (~mask2)|mask4;
      mask3 = ~mask;
      mask4 = ~mask2;
      LSHIFT(src_hi, src_lo, dec_sa+32, &mask5, &dst_hi, &dst_lo);
      dst_lo = (dst_lo & mask3) | (tar_lo & mask);
      dst_hi = (dst_hi & mask4) | (tar_hi & mask2);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 007:/* DINS */
    if(inst64_en){
      LSHIFT(0, 1, dec_rd+1, &mask5, &mask2, &mask);
      ISUB64(mask2, mask, 0, 1, &mask2, &mask, &cy);
      LSHIFT(0, 1, dec_sa, &mask5, &mask4, &mask3);
      ISUB64(mask4, mask3, 0, 1, &mask4, &mask3, &cy);
      mask  = (~mask) |mask3;
      mask2 = (~mask2)|mask4;
      mask3 = ~mask;
      mask4 = ~mask2;
      LSHIFT(src_hi, src_lo, dec_sa, &mask5, &dst_hi, &dst_lo);
      dst_lo = (dst_lo & mask3) | (tar_lo & mask);
      dst_hi = (dst_hi & mask4) | (tar_hi & mask2);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 040:/* BSHFL */
    switch(dec_sa){
    case 002:/* SWBH */
      dst_lo = (((tar_lo>>16)&0xff)<<24)|
               (((tar_lo>>24)&0xff)<<16)|
               (((tar_lo>> 0)&0xff)<< 8)|
               (((tar_lo>> 8)&0xff)<< 0);
      dst_hi = sign_extend(dst_lo);
      set_gr(dec_rd, dst_hi, dst_lo);
      break;
    case 020:/* SEB */
      if(tar_lo&0x80){
        dst_lo = tar_lo | 0xffffff00;
        dst_hi = 0xffffffff;
      }
      else{
        dst_lo = tar_lo & 0x000000ff;
        dst_hi = 0x00000000;
      }
      set_gr(dec_rd, dst_hi, dst_lo);
      break;     
    case 030:/* SEH */
      if(tar_lo&0x8000){
        dst_lo = tar_lo | 0xffff0000;
        dst_hi = 0xffffffff;
      }
      else{
        dst_lo = tar_lo & 0x0000ffff;
        dst_hi = 0x00000000;
      }
      set_gr(dec_rd, dst_hi, dst_lo);
      break;
    default:
      except = Exc_RI;
    }
    break;
  case 044:/* DBSHFL */
    switch(dec_sa){
    case 002:/* SDBH */
      if(inst64_en){
        dst_lo = (((tar_lo>>16)&0xff)<<24)|
                 (((tar_lo>>24)&0xff)<<16)|
                 (((tar_lo>> 0)&0xff)<< 8)|
                 (((tar_lo>> 8)&0xff)<< 0);
        dst_hi = (((tar_hi>>16)&0xff)<<24)|
                 (((tar_hi>>24)&0xff)<<16)|
                 (((tar_hi>> 0)&0xff)<< 8)|
                 (((tar_hi>> 8)&0xff)<< 0);
        set_gr(dec_rd, dst_hi, dst_lo);
      }
      else{
        except = Exc_RI;
      }
      break;
    case 005:/* DSHD */
      if(inst64_en){
        dst_lo = (((tar_hi>> 0)&0xffff)<<16)|
                 (((tar_hi>>16)&0xffff)<< 0);
        dst_hi = (((tar_lo>> 0)&0xffff)<<16)|
                 (((tar_lo>>16)&0xffff)<< 0);
        set_gr(dec_rd, dst_hi, dst_lo);
      }
      else{
        except = Exc_RI;
      }
      break;
    default:
      except = Exc_RI;
    }
    break;
  default:
    except = Exc_RI;
  }
}

static void execute32(void)
{
  ulong inst;
  ulong link_hi, link_lo;
  ulong base_hi, base_lo;
  ulong immd_hi, immd_lo;
  ulong data_hi, data_lo;
  int   cy, cy2;
  int   ssign, tsign;
  int   lt, eq;

  except    = Exc_NONE;
  br_inst   = 0;
  br_likely = 0;
  no_ds_br  = 0;

  if((PCL&3L)==2L){
    except = Exc_AdEL;
    set_badva(PCH, PCL);
    return;
  }

  inst = get_icache(PCH,PCL,&except);

  if(except!=Exc_NONE){ /* TLBL or ADEL */
    set_badva(PCH, PCL);
    return;
  }

  brk_inst = inst;
  dec_op   = (inst>>26)&077;
  dec_rs   = (inst>>21)&037;
  dec_rt   = (inst>>16)&037;
  dec_rd   = (inst>>11)&037;
  dec_sa   = (inst>> 6)&037;
  dec_fn   = (inst    )&077;
  dec_imm  = (inst    )&0xffff;

  get_gr(dec_rs, &src_hi, &src_lo);
  get_gr(dec_rt, &tar_hi, &tar_lo);

  switch(dec_op){
  case 000:/* SPECIAL */
    exec_special();
    break;
  case 001:/* REGIMM */
    exec_regimm();
    break;
  case 002:/* J */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    target_hi  = base_hi;
    target_lo  = (base_lo&0xf0000000L)|((inst&0x03ffffffL)<<2);
    br_cond    = 1;
    br_likely  = 0;
    br_inst = 4;
    break;
  case 003:/* JAL */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    IADD64(PCH, PCL, 0L, 8L, &link_hi, &link_lo, &cy);
    target_hi  = base_hi;
    target_lo  = (base_lo&0xf0000000L)|((inst&0x03ffffffL)<<2);
    br_cond    = 1;
    br_likely  = 0;
    br_inst = 4;
    set_gr(31, link_hi, link_lo);
    break;
  case 004:/* BEQ */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)==0L);
    br_likely  = 0;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 005:/* BNE */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)!=0L);
    br_likely  = 0;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 006:/* BLEZ */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)==0L)||((dst_hi&0x80000000L)!=0L);
    br_likely  = 0;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 007:/* BGTZ */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)!=0L)&&((dst_hi&0x80000000L)==0L);
    br_likely  = 0;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 010:/* ADDI */
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xffff0000L|dec_imm;
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm;
    }
    IADD64(sign_extend(src_lo), src_lo, sign_extend(immd_lo), immd_lo, &dst_hi, &dst_lo, &cy);
    cy  = ((dst_lo&0x80000000L)!=0);
    cy2 = dst_hi&1L;
    if(cy==cy2){
      set_gr(dec_rt, sign_extend(dst_lo), dst_lo);
    }
    else{
      except = Exc_Ov;
    }
    break;
  case 011:/* ADDIU */
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xffff0000L|dec_imm;
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm;
    }
    IADD64(src_hi, src_lo, immd_hi, immd_lo, &dst_hi, &dst_lo, &cy);
    set_gr(dec_rt, sign_extend(dst_lo), dst_lo);
    break;
  case 012:/* SLTI */
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xffff0000L|dec_imm;
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm;
    }
    CMP64(src_hi, src_lo, immd_hi, immd_lo, &lt, &eq);
    if(lt&&(!eq))
      set_gr(dec_rt, 0L, 1L);
    else
      set_gr(dec_rt, 0L, 0L);
    break;
  case 013:/* SLTIU */
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xffff0000L|dec_imm;
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm;
    }
    UCMP64(src_hi, src_lo, immd_hi, immd_lo, &lt, &eq);
    if(lt&&(!eq))
      set_gr(dec_rt, 0L, 1L);
    else
      set_gr(dec_rt, 0L, 0L);
    break;
  case 014:/* ANDI */
    dst_hi = 0;
    dst_lo = src_lo & (ulong)dec_imm;
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 015:/* ORI */
    dst_hi = src_hi;
    dst_lo = src_lo | (ulong)dec_imm;
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 016:/* XORI */
    dst_hi = src_hi;
    dst_lo = src_lo ^ (ulong)dec_imm;
    set_gr(dec_rt, dst_hi, dst_lo);
    break;
  case 017:/* LUI */
    dst_lo = dec_imm<<16;
    set_gr(dec_rt, sign_extend(dst_lo), dst_lo);
    break;
  case 020:/* COP0 */
    exec_cop0();
    break;
  case 021:/* COP1 */
    exec_cop1();
    break;
  case 022:/* COP2 */
    except = Exc_CpU;
    break;
  case 023:/* COP1X */
    exec_cop1x();
    break;
  case 024:/* BEQL */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)==0L);
    br_likely  = 1;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 025:/* BNEL */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)!=0L);
    br_likely  = 1;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 026:/* BLEZL */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)==0L)||((dst_hi&0x80000000L)!=0L);
    br_likely  = 1;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 027:/* BGTZL */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    if(dec_imm&0x8000){
      immd_hi = 0xffffffffL;
      immd_lo = 0xfffc0000L|(dec_imm<<2);
    }
    else{
      immd_hi = 0L;
      immd_lo = dec_imm<<2;
    }
    IADD64(base_hi, base_lo, immd_hi, immd_lo, &target_hi, &target_lo, &cy);
    ISUB64(src_hi, src_lo, tar_hi, tar_lo, &dst_hi, &dst_lo, &cy);
    br_cond    = ((dst_hi|dst_lo)!=0L)&&((dst_hi&0x80000000L)==0L);
    br_likely  = 1;
    br_inst = 4;
    bt_refer(target_lo,br_cond);
    break;
  case 030:/* DADDI */
    if(inst64_en){
      if(dec_imm&0x8000){
	immd_hi = 0xffffffffL;
	immd_lo = 0xffff0000L|dec_imm;
	tsign = 1;
      }
      else{
	immd_hi = 0L;
	immd_lo = dec_imm;
	tsign = 0;
      }
      ssign = ((src_hi&0x80000000L)!=0L);
      IADD64(src_hi, src_lo, immd_hi, immd_lo, &dst_hi, &dst_lo, &cy);
      cy2 = ((dst_hi&0x80000000L)!=0);
      cy  = (cy+ssign+tsign)&1;
      if(cy==cy2){
	set_gr(dec_rt, dst_hi, dst_lo);
      }
      else{
	except = Exc_Ov;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 031:/* DADDIU */
    if(inst64_en){
      if(dec_imm&0x8000){
	immd_hi = 0xffffffffL;
	immd_lo = 0xffff0000L|dec_imm;
      }
      else{
	immd_hi = 0L;
	immd_lo = dec_imm;
      }
      IADD64(src_hi, src_lo, immd_hi, immd_lo, &dst_hi, &dst_lo, &cy);
      set_gr(dec_rt, dst_hi, dst_lo);
    }
    else{
      except = Exc_RI;
    }
    break;
  case 032:/* LDL */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);/* no error */
	switch(base_lo&0x7){
	case 0:
	  dst_hi = (data_lo<<24)|(tar_hi&0x00ffffffL);
	  dst_lo =                tar_lo;
	  break;
	case 1:
	  dst_hi = (data_lo<<16)|(tar_hi&0x0000ffffL);
	  dst_lo =                tar_lo;
	  break;
	case 2:
	  dst_hi = (data_lo<< 8)|(tar_hi&0x000000ffL);
	  dst_lo =                tar_lo;
	  break;
	case 3:
	  dst_hi = data_lo;
	  dst_lo = tar_lo;
	  break;
	case 4:
	  dst_hi = (data_lo>> 8)|(data_hi<<24);
	  dst_lo = (data_lo<<24)|(tar_lo&0x00ffffffL);
	  break;
	case 5:
	  dst_hi = (data_lo>>16)|(data_hi<<16);
	  dst_lo = (data_lo<<16)|(tar_lo&0x0000ffffL);
	  break;
	case 6:
	  dst_hi = (data_lo>>24)|(data_hi<< 8);
	  dst_lo = (data_lo<< 8)|(tar_lo&0x000000ffL);
	  break;
	case 7:
	  dst_hi = data_hi;
	  dst_lo = data_lo;
	  break;
	}
	set_gr(dec_rt, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 033:/* LDR */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);/* no error */
	switch(base_lo&0x7){
	case 0:
	  dst_hi = data_hi;
	  dst_lo = data_lo;
	  break;
	case 1:
	  dst_hi = (data_hi>> 8)|(tar_hi&0xff000000L);
	  dst_lo = (data_hi<<24)|(data_lo>> 8);
	  break;
	case 2:
	  dst_hi = (data_hi>>16)|(tar_hi&0xffff0000L);
	  dst_lo = (data_hi<<16)|(data_lo>>16);
	  break;
	case 3:
	  dst_hi = (data_hi>>24)|(tar_hi&0xffffff00L);
	  dst_lo = (data_hi<< 8)|(data_lo>>24);
	  break;
	case 4:
	  dst_hi = tar_hi;
	  dst_lo = data_hi;
	  break;
	case 5:
	  dst_hi = tar_hi;
	  dst_lo = (data_hi>> 8)|(tar_lo&0xff000000L);
	  break;
	case 6:
	  dst_hi = tar_hi;
	  dst_lo = (data_hi>>16)|(tar_lo&0xffff0000L);
	  break;
	case 7:
	  dst_hi = tar_hi;
	  dst_lo = (data_hi>>24)|(tar_lo&0xffffff00L);
	  break;
	}
	set_gr(dec_rt, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 034:/* SPECIAL2 */
    exec_special2();
    break;
  case 035:/* JALX */
    IADD64(PCH, PCL, 0L, 4L, &base_hi, &base_lo, &cy);
    IADD64(PCH, PCL, 0L, 8L, &link_hi, &link_lo, &cy);
    target_hi  = base_hi;
    target_lo  = (base_lo&0xf0000000L)|((inst&0x03ffffffL)<<2);
    br_cond    = 1;
    br_likely  = 0;
    br_inst = 4;
    target_lo ^= 1L;
    set_gr(31, link_hi, link_lo);
    break;
  case 036:/* MDMX */
    if(mips32_en)
      except = Exc_MDMX;
    else
      except = Exc_RI;
    break;
  case 037:/* SPECIAL3 */
    if(ver2_en)
      exec_special3();
    else
      except = Exc_RI;
    break;
  case 040:/* LB */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	break;
      case 1:
	data_lo = data_lo>> 8;
	break;
      case 2:
	data_lo = data_lo>>16;
	break;
      case 3:
	data_lo = data_lo>>24;
	break;
      }
      if(data_lo&0x80){
	dst_lo = 0xffffff00L|(data_lo&0x000000ffL);
	dst_hi = 0xffffffffL;
      }
      else{
	dst_lo = (data_lo&0x000000ffL);
	dst_hi = 0L;
      }
      set_gr(dec_rt, dst_hi, dst_lo);
      LLbit = 0;
    }
    break;
  case 041:/* LH */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    if(base_lo&0x1){
      except = Exc_AdEL;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	if(base_lo&0x3){
	  data_lo = data_lo>>16;
	}
	if(data_lo&0x8000){
	  dst_lo = 0xffff0000L|(data_lo&0x0000ffffL);
	  dst_hi = 0xffffffffL;
	}
	else{
	  dst_lo = (data_lo&0x0000ffffL);
	  dst_hi = 0L;
	}
      	set_gr(dec_rt, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 042:/* LWL */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	dst_lo = (data_lo<<24)|(tar_lo&0x00ffffffL);
	break;
      case 1:
	dst_lo = (data_lo<<16)|(tar_lo&0x0000ffffL);
	break;
      case 2:
	dst_lo = (data_lo<< 8)|(tar_lo&0x000000ffL);
	break;
      case 3:
	dst_lo = data_lo;
	break;
      }
      if(dst_lo&0x80000000L){
	dst_hi = 0xffffffffL;
      }
      else{
	dst_hi = 0L;
      }
      set_gr(dec_rt, dst_hi, dst_lo);
      LLbit = 0;
    }
    break;
  case 043:/* LW */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    if(base_lo&0x3){
      except = Exc_AdEL;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	dst_lo = data_lo;
	if(data_lo&0x80000000L){
	  dst_hi = 0xffffffffL;
	}
	else{
	  dst_hi = 0L;
	}
      	set_gr(dec_rt, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 044:/* LBU */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	break;
      case 1:
	data_lo = data_lo>> 8;
	break;
      case 2:
	data_lo = data_lo>>16;
	break;
      case 3:
	data_lo = data_lo>>24;
	break;
      }
      dst_lo = data_lo&0x000000ffL;
      dst_hi = 0L;
      set_gr(dec_rt, dst_hi, dst_lo);
      LLbit = 0;
    }
    break;
  case 045:/* LHU */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    if(base_lo&0x1){
      except = Exc_AdEL;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	if(base_lo&0x3){
	  data_lo = data_lo>>16;
	}
	dst_lo = data_lo&0x0000ffffL;
	dst_hi = 0L;
      	set_gr(dec_rt, dst_hi, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 046:/* LWR */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
 	dst_lo = data_lo;
	break;
      case 1:
	dst_lo = (data_lo>> 8)|(tar_lo&0xff000000L);
	break;
      case 2:
	dst_lo = (data_lo>>16)|(tar_lo&0xffff0000L);
	break;
      case 3:
	dst_lo = (data_lo>>24)|(tar_lo&0xffffff00L);
	break;
      }
      if(dst_lo&0x80000000L){
	dst_hi = 0xffffffffL;
      }
      else{
	dst_hi = 0L;
      }
      if(ty_cpu==R4300)
	get_gr(dec_rt, &dst_hi, &data_lo);
      set_gr(dec_rt, dst_hi, dst_lo);
      LLbit = 0;
    }
    break;
  case 047:/* LWU */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&0x3){
	except = Exc_AdEL;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
	  dst_lo = data_lo;
	  dst_hi = 0L;
	  set_gr(dec_rt, dst_hi, dst_lo);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 050:/* SB */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	dst_lo = (tar_lo&0x000000ffL)|(data_lo&0xffffff00L);
	break;
      case 1:
	dst_lo = ((tar_lo&0x000000ffL)<<8)|(data_lo&0xffff00ffL);
	break;
      case 2:
	dst_lo = ((tar_lo&0x000000ffL)<<16)|(data_lo&0xff00ffffL);
	break;
      case 3:
	dst_lo = ((tar_lo&0x000000ffL)<<24)|(data_lo&0x00ffffffL);
	break;
      }
      put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
      LLbit = 0;
    }
    break;
  case 051:/* SH */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    if(base_lo&0x1){
      except = Exc_AdES;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	switch(base_lo&0x2){
	case 0:
	  dst_lo = (tar_lo&0x0000ffffL)|(data_lo&0xffff0000L);
	  break;
	case 2:
	  dst_lo = ((tar_lo&0x0000ffffL)<<16)|(data_lo&0x0000ffffL);
	  break;
	}
	put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 052:/* SWL */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	dst_lo = (tar_lo>>24)|(data_lo&0xffffff00L);
	break;
      case 1:
	dst_lo = (tar_lo>>16)|(data_lo&0xffff0000L);
	break;
      case 2:
	dst_lo = (tar_lo>> 8)|(data_lo&0xff000000L);
	break;
      case 3:
	dst_lo = tar_lo;
	break;
      }
      put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
      LLbit = 0;
    }
    break;
  case 053:/* SW */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    if(base_lo&0x3){
      except = Exc_AdES;
      set_badva(base_hi, base_lo);
    }
    else{
      data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	dst_lo = tar_lo;
	put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
	LLbit = 0;
      }
    }
    break;
  case 054:/* SDL */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
	switch(base_lo&0x7){
	case 0:
	  dst_lo = (tar_hi>>24)|(data_lo&0xffffff00L);
	  dst_hi = data_hi;
	  break;
	case 1:
	  dst_lo = (tar_hi>>16)|(data_lo&0xffff0000L);
	  dst_hi = data_hi;
	  break;
	case 2:
	  dst_lo = (tar_hi>> 8)|(data_lo&0xff000000L);
	  dst_hi = data_hi;
	  break;
	case 3:
	  dst_lo = tar_hi;
	  dst_hi = data_hi;
	  break;
	case 4:
	  dst_lo = (tar_hi<< 8)|(tar_lo>>24);
	  dst_hi = (tar_hi>>24)|(data_hi&0xffffff00L);
	  break;
	case 5:
	  dst_lo = (tar_hi<<16)|(tar_lo>>16);
	  dst_hi = (tar_hi>>16)|(data_hi&0xffff0000L);
	  break;
	case 6:
	  dst_lo = (tar_hi<<24)|(tar_lo>> 8);
	  dst_hi = (tar_hi>> 8)|(data_hi&0xff000000L);
	  break;
	case 7:
	  dst_lo = tar_lo;
	  dst_hi = tar_hi;
	  break;
	}
	put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
	put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
	LLbit = 0;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 055:/* SDR */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
      if(except!=Exc_NONE)
	set_badva(base_hi, base_lo);
      if(except==Exc_NONE){
	data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
	switch(base_lo&0x7){
	case 0:
	  dst_lo = tar_lo;
	  dst_hi = tar_hi;
	  break;
	case 1:
	  dst_lo = (tar_lo<< 8)|(data_lo&0x000000ffL);
	  dst_hi = (tar_hi<< 8)|(tar_lo>>24);
	  break;
	case 2:
	  dst_lo = (tar_lo<<16)|(data_lo&0x0000ffffL);
	  dst_hi = (tar_hi<<16)|(tar_lo>>16);
	  break;
	case 3:
	  dst_lo = (tar_lo<<24)|(data_lo&0x00ffffffL);
	  dst_hi = (tar_hi<<24)|(tar_lo>> 8);
	  break;
	case 4:
	  dst_lo = data_lo;
	  dst_hi = tar_lo;
	  break;
	case 5:
	  dst_lo = data_lo;
	  dst_hi = (tar_lo<< 8)|(data_hi&0x000000ffL);
	  break;
	case 6:
	  dst_lo = data_lo;
	  dst_hi = (tar_lo<<16)|(data_hi&0x0000ffffL);
	  break;
	case 7:
	  dst_lo = data_lo;
	  dst_hi = (tar_lo<<24)|(data_hi&0x00ffffffL);
	  break;
	}
	put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
	put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
	LLbit = 0;
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 056:/* SWR */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
    if(except!=Exc_NONE)
      set_badva(base_hi, base_lo);
    if(except==Exc_NONE){
      switch(base_lo&0x3){
      case 0:
	dst_lo = tar_lo;
	break;
      case 1:
	dst_lo = (tar_lo<< 8)|(data_lo&0x000000ffL);
	break;
      case 2:
	dst_lo = (tar_lo<<16)|(data_lo&0x0000ffffL);
	break;
      case 3:
	dst_lo = (tar_lo<<24)|(data_lo&0x00ffffffL);
	break;
      }
      put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
      LLbit = 0;
    }
    break;
  case 057:/* CACHE */
    ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
    exec_cache(dec_rt, base_hi, base_lo);
    break;
  case 060:/* LL */
    if(ty_llsc||(ty_cpu==R4300)){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&0x3){
        except = Exc_AdEL;
        set_badva(base_hi, base_lo);
      }
      else{
        data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
        if(except!=Exc_NONE)
          set_badva(base_hi, base_lo);
        if(except==Exc_NONE){
          dst_lo = data_lo;
          if(data_lo&0x80000000L){
            dst_hi = 0xffffffffL;
          }
          else{
            dst_hi = 0L;
          }
          set_gr(dec_rt, sign_extend(dst_lo), dst_lo);
          LLbit = 1;
	  if(ty_cpu==R4300)
	    mtc0(C0_LLAddr, get_lastpa()>>4);
	  else
	    mtc0(C0_LLAddr, get_lastpa());
        }
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 061:/* LWC1 */
    if(cp1_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&0x3){
        except = Exc_AdEL;
        set_badva(base_hi, base_lo);
      }
      else{
        data_lo = get_dcachel(base_hi, base_lo&0xfffffffcL, &except);
        if(except!=Exc_NONE)
          set_badva(base_hi, base_lo);
        if(except==Exc_NONE){
          dst_lo = data_lo;
          if(data_lo&0x80000000L){
            dst_hi = 0xffffffffL;
          }
          else{
            dst_hi = 0L;
          }
	  if(fr_bit)
	    set_fr64(dec_rt, dst_hi, dst_lo);
	  else
	    set_fr32(dec_rt, dst_lo);
	  LLbit = 0;
        }
      }
    }
    else{
      except = Exc_CpU;
    }
    break;
  case 062:/* LWC2 */
    except = Exc_CpU;
    break;
  case 063:/* PREF */
    if(!mips4_en){
      except = Exc_RI;
    }
    break;
  case 064:/* LLD */
    if(inst64_en&&(ty_llsc||(ty_cpu==R4300))){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
	except = Exc_AdEL;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
	  data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);
	  dst_lo = data_lo;
	  dst_hi = data_hi;
	  set_gr(dec_rt, dst_hi, dst_lo);
          LLbit = 1;
	  if(ty_cpu==R4300)
	    mtc0(C0_LLAddr, (get_lastpa()&0xfffffff8L)>>4);
	  else
	    mtc0(C0_LLAddr, (get_lastpa()&0xfffffff8L));
	}
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 065:/* LDC1 */
    if(cp1_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
	except = Exc_AdEL;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
	  data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);
	  dst_lo = data_lo;
	  dst_hi = data_hi;
	  set_fr64(dec_rt, dst_hi, dst_lo);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_CpU;
    }
    break;
  case 066:/* LDC2 */
    if(inst64_en){
      except = Exc_CpU;
    }
    else{
      except = Exc_RI;
    }
    break;
  case 067:/* LD */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
	except = Exc_AdEL;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcachel(base_hi, base_lo&0xfffffff8L, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
	  data_hi = get_dcachel(base_hi, (base_lo&0xfffffff8L)+4, &except);
	  dst_lo = data_lo;
	  dst_hi = data_hi;
	  set_gr(dec_rt, dst_hi, dst_lo);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 070:/* SC */
    if(ty_llsc||(ty_cpu==R4300)){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&0x3){
        except = Exc_AdES;
        set_badva(base_hi, base_lo);
      }
      else{
        data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
        if(except!=Exc_NONE)
          set_badva(base_hi, base_lo);
        if(except==Exc_NONE){
          if(LLbit){
            dst_lo = tar_lo;
            put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
          }
          set_gr(dec_rt, 0L, (ulong)LLbit);
	  LLbit = 0;
        }
      }
    }
    else{
      except =Exc_RI;
    }
    break;
  case 071:/* SWC1 */
    if(cp1_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&0x3){
        except = Exc_AdES;
        set_badva(base_hi, base_lo);
      }
      else{
        data_lo = get_dcaches(base_hi, base_lo&0xfffffffcL, &except);
        if(except!=Exc_NONE)
          set_badva(base_hi, base_lo);
        if(except==Exc_NONE){
          dst_lo = get_fr32(dec_rt);
          put_dcache(base_hi, base_lo&0xfffffffcL, dst_lo);
	  LLbit = 0;
        }
      }
    }
    else{
      except =Exc_CpU;
    }
    break;
  case 072:/* SWC2 */
    except = Exc_CpU;
    break;
  case 074:/* SCD */
    if(inst64_en&&(ty_llsc||(ty_cpu==R4300))){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
	except = Exc_AdES;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
          if(LLbit){
            data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
            dst_lo = tar_lo;
            dst_hi = tar_hi;
	    put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
	    put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
          }
          set_gr(dec_rt, 0L, (ulong)LLbit);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  case 075:/* SDC1 */
    if(cp1_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
	except = Exc_AdES;
	set_badva(base_hi, base_lo);
      }
      else{
	data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
	if(except!=Exc_NONE)
	  set_badva(base_hi, base_lo);
	if(except==Exc_NONE){
	  data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
          get_fr64(dec_rt, &dst_hi, &dst_lo);
	  put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
	  put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_CpU;
    }
    break;
  case 076:/* SDC2 */
    if(inst64_en){
      except = Exc_CpU;
    }
    else{
      except =Exc_RI;
    }
    break;
  case 077:/* SD */
    if(inst64_en){
      ea_gen(src_hi, src_lo, dec_imm, &base_hi, &base_lo);
      if(base_lo&7L){
    	except = Exc_AdES;
     	set_badva(base_hi, base_lo);
      }
      else{
     	data_lo = get_dcaches(base_hi, base_lo&0xfffffff8L, &except);
       	if(except!=Exc_NONE)
     	  set_badva(base_hi, base_lo);
       	if(except==Exc_NONE){
    	  data_hi = get_dcaches(base_hi, (base_lo&0xfffffff8L)+4, &except);
          dst_lo = tar_lo;
	  dst_hi = tar_hi;
	  put_dcache(base_hi, (base_lo&0xfffffff8L)+0, dst_lo);
	  put_dcache(base_hi, (base_lo&0xfffffff8L)+4, dst_hi);
	  LLbit = 0;
	}
      }
    }
    else{
      except = Exc_RI;
    }
    break;
  default:/* Reserved Instruction */
    except = Exc_RI;
  }
}

static int get_bev(void)
{
  ulong v;

  v = mfc0(C0_SR);
  return (v&(ulong)(1<<22))!=0L;
}

static int get_exl(void)
{
  ulong v;

  v = mfc0(C0_SR);
  return (v&(ulong)(1<<1))!=0L;
}

static void gen_vector(ulong *hi, ulong *lo)
{
  if(get_bev()){
    if(((except==Exc_TLBML)||(except==Exc_TLBMS))
       && (get_exl()==0)){ /* TLB Miss */
      if(addr64){
	*hi = 0xffffffffL;
	*lo = 0xbfc00280L;
      }
      else{
	*hi = 0xffffffffL;
	*lo = 0xbfc00200L;
      }
    }
    else{
      *hi = 0xffffffffL;
      if(mips32_en && (except==Exc_Int) && (get_cause()&(1<<23)))
        *lo = 0xbfc00400L;
      else
        *lo = 0xbfc00380L;
    }
  }
  else{
    if(((except==Exc_TLBML)||(except==Exc_TLBMS))
       && (get_exl()==0)){ /* TLB Miss */
      if(addr64){
	*hi = 0xffffffffL;
	*lo = 0x80000080L;
      }
      else{
	*hi = 0xffffffffL;
	*lo = 0x80000000L;
      }
    }
    else{
      *hi = 0xffffffffL;
      if(mips32_en && (except==Exc_Int) && (get_cause()&(1<<23)))
        *lo = 0x80000200L;
      else
        *lo = 0x80000180L;
    }
  }
}

static void set_excode(void)
{
  ulong v;
  int   e;

  if(except==Exc_TLBML)
    e = Exc_TLBL;
  else if(except==Exc_TLBMS)
    e = Exc_TLBS;
  else
    e = except;
  v = get_cause();
  v &= 0x48c0ff00L;
  if(delay_slot)
    v |= 0x80000000L;          /* BD */
  if(except==Exc_CpU)
    v |= (brk_inst&0x30000000L); /* CE */
  v |= (ulong)e;               /* ExCode */
  set_cause(v);
}

static void set_epc(void)
{
  if(!get_exl()){
    dmtc0(C0_EPC, PCH, PCL);
  }
}

static void set_exl(void)
{
  ulong v;

  v = mfc0(C0_SR);
  v |= 2L;
  mtc0(C0_SR, v);
}

static void gen_nextpc(void)
{
  ulong next_hi, next_lo;
  int   cy;

  PCH_p = PCH;
  PCL_p = PCL;
/*
printf("PC=%08x %08x\n",PCH,PCL);
printf("except=%d\n",except);
printf("target=%08x %08x\n",target_hi,target_lo);
printf("intr=%d\n",int_check());
*/
  if(watch_pend){
	cy = get_exl();
	if((!cy)&&(except==Exc_NONE)){
	  except = Exc_WATCH;
	  watch_pend = 0;
	}
  }
  if(except!=Exc_NONE){
    if(int_check())
      except = Exc_Int;
  }
  if(delay_slot){
    if(br_cond){ /* taken */
      if(except!=Exc_NONE){
	ISUB64(PCH, PCL, 0L, (ulong)delay_slot, &PCH, &PCL, &cy);
	gen_vector(&next_hi, &next_lo);
	set_excode();
	set_epc();
	set_exl();
      }
      else{
	next_hi = target_hi;
	next_lo = target_lo;
      }
    }
    else{ /* no taken */
      if(except!=Exc_NONE){
	ISUB64(PCH, PCL, 0L, (ulong)delay_slot, &PCH, &PCL, &cy);
	gen_vector(&next_hi, &next_lo);
	set_excode();
	set_epc();
	set_exl();
      }
      else{
	if(PCL&1L)
	  IADD64(PCH, PCL, 0L, 2L, &next_hi, &next_lo, &cy);
	else
	  IADD64(PCH, PCL, 0L, 4L, &next_hi, &next_lo, &cy);
      }
    }
  }
  else{
    if(except!=Exc_NONE){
      if(extend_flag){
	ISUB64(PCH, PCL, 0L, 2L, &PCH, &PCL, &cy);
	extend_flag = 0;
      }
      gen_vector(&next_hi, &next_lo);
      set_excode();
      set_epc();
      set_exl();
    }
    else if(no_ds_br){
      next_hi = target_hi;
      next_lo = target_lo;
    }
    else if(br_likely&&(br_cond==0)){ /* kill */
      IADD64(PCH, PCL, 0L, 8L, &next_hi, &next_lo, &cy);
      br_inst = 0;
    }
    else{
      if(PCL&1L)
	IADD64(PCH, PCL, 0L, 2L, &next_hi, &next_lo, &cy);
      else
	IADD64(PCH, PCL, 0L, 4L, &next_hi, &next_lo, &cy);
    }
  }
  PCH = next_hi;
  PCL = next_lo;
  delay_slot = br_inst;
}

static void get_enable(void)
{
  ulong v;
  int   exl, erl, ksu, kx, sx, ux, cu0, cu1, cu3, px;

  v = mfc0(C0_SR);
  exl = (v>>1)&1;
  erl = (v>>2)&1;
  ksu = (v>>3)&3;
  ux  = (v>>5)&1;
  sx  = (v>>6)&1;
  kx  = (v>>7)&1;
  cu0 = (v>>28)&1;
  cu1 = (v>>29)&1;
  cu3 = (v>>31)&1;
  px  = (v>>23)&mips32_en&1;

  fr_bit = ((v&(ulong)(1<<26))!=0L);
  cp0_en    = cu0||exl||erl||(ksu==0);
  cp1_en    = cu1;
  mips4_en  = (mips32_en)? 1 : (!no_mips4)&&(cu3||(ksu!=2)||exl||erl);
  addr64    =   ((ksu==2)&&ux) /* user 64bit */
              ||((ksu==1)&&sx) /* supervisoer 64bit */
              ||((ksu==0)&&kx) /* kernel 64bit */
              ||((exl||erl)&&kx); /* in handler */
  inst64_en =   ((ksu==2)&&(ux||px)) /* user 64bit */
              ||((ksu==1)&&(sx||ver2_en)) /* supervisoer 64bit */
              || (ksu==0)      /* kernel */
              ||exl||erl;      /* in handler */
}

void execute(void)
{
  get_enable();
  if(PCL&1L)
    execute16(PCH, PCL);
  else
    execute32();
  gen_nextpc();
  update_random();
  update_count();
}

void cpu_dump(void)
{
  printf("[cpu]\n");
  printf("pc:%08x%08x 01:%08x%08x 02:%08x%08x 03:%08x%08x\n",
	 PCH_p,PCL_p,       GR[1][0],GR[1][1],
	 GR[2][0],GR[2][1], GR[3][0],GR[3][1]);
  printf("04:%08x%08x 05:%08x%08x 06:%08x%08x 07:%08x%08x\n",
	 GR[4][0],GR[4][1], GR[5][0],GR[5][1],
	 GR[6][0],GR[6][1], GR[7][0],GR[7][1]);
  printf("08:%08x%08x 09:%08x%08x 10:%08x%08x 11:%08x%08x\n",
	 GR[8][0],GR[8][1], GR[9][0],GR[9][1],
	 GR[10][0],GR[10][1], GR[11][0],GR[11][1]);
  printf("12:%08x%08x 13:%08x%08x 14:%08x%08x 15:%08x%08x\n",
	 GR[12][0],GR[12][1], GR[13][0],GR[13][1],
	 GR[14][0],GR[14][1], GR[15][0],GR[15][1]);
  printf("16:%08x%08x 17:%08x%08x 18:%08x%08x 19:%08x%08x\n",
	 GR[16][0],GR[16][1], GR[17][0],GR[17][1],
	 GR[18][0],GR[18][1], GR[19][0],GR[19][1]);
  printf("20:%08x%08x 21:%08x%08x 22:%08x%08x 23:%08x%08x\n",
	 GR[20][0],GR[20][1], GR[21][0],GR[21][1],
	 GR[22][0],GR[22][1], GR[23][0],GR[23][1]);
  printf("24:%08x%08x 25:%08x%08x 26:%08x%08x 27:%08x%08x\n",
	 GR[24][0],GR[24][1], GR[25][0],GR[25][1],
	 GR[26][0],GR[26][1], GR[27][0],GR[27][1]);
  printf("28:%08x%08x 29:%08x%08x 30:%08x%08x 31:%08x%08x\n",
	 GR[28][0],GR[28][1], GR[29][0],GR[29][1],
	 GR[30][0],GR[30][1], GR[31][0],GR[31][1]);
  printf("hi:%08x%08x lo:%08x%08x\n",
	 HIHI,HILO, LOHI,LOLO);
}
void fpu_dump(void)
{
  printf("[fpu]\n");
  printf("00:%08x%08x 01:%08x%08x 02:%08x%08x 03:%08x%08x\n",
	 FR[0][1],FR[0][0], FR[1][1],FR[1][0],
	 FR[2][1],FR[2][0], FR[3][1],FR[3][0]);
  printf("04:%08x%08x 05:%08x%08x 06:%08x%08x 07:%08x%08x\n",
	 FR[4][1],FR[4][0], FR[5][1],FR[5][0],
	 FR[6][1],FR[6][0], FR[7][1],FR[7][0]);
  printf("08:%08x%08x 09:%08x%08x 10:%08x%08x 11:%08x%08x\n",
	 FR[8][1],FR[8][0], FR[9][1],FR[9][0],
	 FR[10][1],FR[10][0], FR[11][1],FR[11][0]);
  printf("12:%08x%08x 13:%08x%08x 14:%08x%08x 15:%08x%08x\n",
	 FR[12][1],FR[12][0], FR[13][1],FR[13][0],
	 FR[14][1],FR[14][0], FR[15][1],FR[15][0]);
  printf("16:%08x%08x 17:%08x%08x 18:%08x%08x 19:%08x%08x\n",
	 FR[16][1],FR[16][0], FR[17][1],FR[17][0],
	 FR[18][1],FR[18][0], FR[19][1],FR[19][0]);
  printf("20:%08x%08x 21:%08x%08x 22:%08x%08x 23:%08x%08x\n",
	 FR[20][1],FR[20][0], FR[21][1],FR[21][0],
	 FR[22][1],FR[22][0], FR[23][1],FR[23][0]);
  printf("24:%08x%08x 25:%08x%08x 26:%08x%08x 27:%08x%08x\n",
	 FR[24][1],FR[24][0], FR[25][1],FR[25][0],
	 FR[26][1],FR[26][0], FR[27][1],FR[27][0]);
  printf("28:%08x%08x 29:%08x%08x 30:%08x%08x 31:%08x%08x\n",
	 FR[28][1],FR[28][0], FR[29][1],FR[29][0],
	 FR[30][1],FR[30][0], FR[31][1],FR[31][0]);
  printf("sr:%08x\n", FCR31);
}
