/*************************************************/
/* All rights reserved by Akira Nakamori. This   */
/* program must be uesed solely for the purpose  */ 
/* for which it was furnished by Akira Nakamori. */
/*                                               */
/*     Copyright (C) Akira Nakamori 2004         */
/*************************************************/
#include <stdio.h>
#include "define.h"

static int   inexact;
static int   abs_neg;
static int   madd_e;
static int   madd_op;

static int   fp_op;
static int   fp_fmt;
static int   fp_rm;
static int   round_inh;
static int   shift_inh;
static int   fp_fs;
static int   fp_cmp;
static int   to_fix;
static int   mul_err;
static int   sqr_err;
static int   operand1;
static int   fp_except;

static int   fs_sgn;
static int   fs_exp;
static ulong fs_manh;
static ulong fs_manl;
static int   fs_qnan;
static int   fs_snan;
static int   fs_zero;
static int   fs_deno;
static int   fs_infi;

static int   ft_sgn;
static int   ft_exp;
static ulong ft_manh;
static ulong ft_manl;
static int   ft_qnan;
static int   ft_snan;
static int   ft_zero;
static int   ft_deno;
static int   ft_infi;

static int   fd_sgn;
static int   fd_exp;
static ulong fd_manh;
static ulong fd_manl;
static int   fd_infi;
static int   fd_zero;

static int   fr_sgn;
static int   fr_exp;
static ulong fr_manh;
static ulong fr_manl;
static int   fr_qnan;
static int   fr_snan;
static int   fr_zero;
static int   fr_deno;
static int   fr_infi;

static int   cvt_e_chk;
static int   cvt_ov;
static int   f_e;
static int   f_v;
static int   f_z;
static int   f_o;
static int   f_u;
static int   f_i;
static int   f_cc;

#define sign_extend(x) ((x&0x80000000)? 0xffffffff : 0)

static void UNPACK(ulong hi, ulong lo, 
                   int *sgn, int *exp, ulong *manh, ulong *manl,
                   int *qnan, int *snan, int *zero, int *deno, int *infi)
{
  ulong tmph, tmpl;
  ulong tth, ttl;
  int nan;

  *sgn  = (fp_fmt == fmtD)? ((hi&0x80000000L)!=0L) :
          (fp_fmt == fmtS)? ((lo&0x80000000L)!=0L) :
          (fp_fmt == fmtL)? ((hi&0x80000000L)!=0L) :
	  (fp_fmt == fmtW)? ((lo&0x80000000L)!=0L) :
          0;
  *exp  = (fp_fmt == fmtD)? ((hi>>20)&0x7ff) :
          (fp_fmt == fmtS)? ((lo>>23)&0xff)  : 
          0;
  tmph  = (fp_fmt == fmtD)? ((hi&0x000fffffL)|((*exp)?(ulong)(1<<20):0L)) :
          (fp_fmt == fmtS)? 0L                                    :
          hi;
  tmpl  = (fp_fmt == fmtD)? lo                                   :
          (fp_fmt == fmtS)? ((lo&0x007fffffL)|((*exp)?(ulong)(1<<23):0L)) :
          lo;
  if((fp_fmt==fmtL)&&(*sgn)){
    INEG64(tmph, tmpl, &tth, &ttl);
    tmph = tth;
    tmpl = ttl;
  }
  else if((fp_fmt==fmtW)&&(*sgn)){
    tmph = 0;
    tmpl = -(long)tmpl;
  }
#if 0
  *manh = (fp_fmt != fmtPS)? ((tmph<<3)|(tmpl>>29)) : tmph;
  *manl = (fp_fmt != fmtPS)? (tmpl<<3)              : tmpl;
#else
  if((fp_fmt==fmtS)||(fp_fmt==fmtD))
    LSHIFT(tmph, tmpl, 3, &ttl, manh, manl);
  else{
    *manh = tmph;
    *manl = tmpl;
  }
#endif
  nan  = (fp_fmt==fmtD)? ((*exp==0x7ff)&&
                          (((*manh&0xff7fffffL)|*manl)!=0L)) :
         (fp_fmt==fmtS)? ((*exp==0xff)&&
                          ((*manl&0xfbffffffL)!=0L)) :
         0;
  *deno = (fp_fmt==fmtD)? ((*exp==0)&&
                           (((*manh&0xff7fffffL)!=0L)||(*manl!=0L))) :
          (fp_fmt==fmtS)? ((*exp==0)&&
                           ((*manl&0xfbffffffL)!=0L)) :
          0;
  *zero = (fp_fmt==fmtD)? ((*exp==0)&&
                           (((*manh&0xff7fffffL)==0L)&&(*manl==0L))):
          (fp_fmt==fmtS)? ((*exp==0)&&
                           ((*manl&0xfbffffffL)==0L)) :
          0;
  *infi = (fp_fmt==fmtD)? ((*exp==0x7ff)&&
                           ((*manh&0xff7fffffL)==0L)&&
                           (*manl==0L)) :
          (fp_fmt==fmtS)? ((*exp==0xff)&&
                           ((*manl&0xfbffffffL)==0L)) :
          0;
  if(nan){
    if(fp_fmt==fmtD){
      if(*manh&(1<<22)){
        *snan = 1;
        *qnan = 0;
      }
      else{
        *snan = 0;
        *qnan = 1;
      }
    }
    else if(fp_fmt==fmtS){
      if(*manl&(1<<25)){
        *snan = 1;
        *qnan = 0;
      }
      else{
        *snan = 0;
        *qnan = 1;
      }
    }
    else{
      *qnan = 0;
      *snan = 0;
    }
  }
  else{
    *qnan = 0;
    *snan = 0;
  }
}

static void NORMALIZE(ulong hi, ulong lo)
{
  int clz;
  ulong dummy;

  if((hi|lo)==0L){
    fd_exp = 0;
    fd_manh = 0L;
    fd_manl = 0L;
    return;
  }
  clz = CLZ64(hi, lo);
  if(fp_fmt == fmtD){
    if(hi&(1<<(56-32))){
      RSHIFTS(hi, lo, 1, &fd_manh, &fd_manl);
      fd_exp++;
    }
    else{
        LSHIFT(hi, lo, clz-8, &dummy, &fd_manh, &fd_manl);
        fd_exp -= (clz-8);
    }
  }
  else{ /* fmtS */
    if(lo&(1<<27)){
      RSHIFTS(hi, lo, 1, &fd_manh, &fd_manl);
      fd_exp++;
    }
    else{
      LSHIFT(hi, lo, clz-37, &dummy, &fd_manh, &fd_manl);
      fd_exp -= (clz-37);
    }
 }
}

static void ROUND(int rm)
{
  ulong adj;
  ulong hi,lo;
  int   dummy;

  inexact = ((fd_manl&7L)!=0);
  switch(rm){
  case rmRN:
    if((fd_manl&4L)==0L)  /* Guard Bit */
      adj = 0L;
    else if(fd_manl&2L)  /* Round Bit */
      adj = 8L;
    else if(fd_manl&1L)  /* Sticky Bit */
      adj = 8L;
    else if(fd_manl&8L)  /* Mantissa LSB */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRZ:
    adj = 0L;
    break;
  case rmRP:
    if(fd_sgn)
      adj = 0;
    else if(fd_manl&4L)  /* Guard Bit */
      adj = 8L;
    else if(fd_manl&2L)  /* Round Bit */
      adj = 8L;
    else if(fd_manl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRM:
    if(fd_sgn==0)
      adj = 0L;
    else if(fd_manl&4L)  /* Guard Bit */
      adj = 8L;
    else if(fd_manl&2L)  /* Round Bit */
      adj = 8L;
    else if(fd_manl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  }
  IADD64(fd_manh, fd_manl, 0L, adj, &hi, &lo, &dummy);
  if(
     ((fp_fmt==fmtD)&&(hi&(1<<24)))
   ||((fp_fmt==fmtS)&&(lo&(1<<27))) ){
    RSHIFTS(hi, lo, 1, &fd_manh, &fd_manl);
    fd_exp++;
  }
  else{
    fd_manh = hi;
    fd_manl = lo;
  }
}

static void REPACK(ulong *fdh, ulong *fdl)
{
  ulong hi,lo;

  hi = (shift_inh==0)? (fd_manh>>3) : fd_manh;
  lo = (shift_inh==0)? ((fd_manl>>3)|(fd_manh<<29)) : fd_manl;
  if(fp_fmt==fmtD){
    hi &= 0x000fffff;
    hi |= ((fd_exp&0x7ff)<<20);
    hi |= (fd_sgn<<31);
    if(fd_infi&&(!f_v)){
      if(fd_sgn){
        hi = 0xfff00000L;
        lo = 0x00000000L;
      }
      else{
        hi = 0x7ff00000L;
        lo = 0x00000000L;
      }
    }
    if(fd_zero&&(!f_v)){
      if(fd_sgn){
        hi = 0x80000000L;
        lo = 0x00000000L;
      }
      else{
        hi = 0x00000000L;
        lo = 0x00000000L;
      }
    }
  }
  else if(fp_fmt==fmtS){
    hi = 0L;
    lo &= 0x007fffffL;
    lo |= ((fd_exp&0xff)<<23);
    lo |= (fd_sgn<<31);
    if(fd_infi&&(!f_v)){
      if(fd_sgn){
        lo = 0xff800000L;
      }
      else{
        lo = 0x7f800000L;
      }
    }
    if(fd_zero&&(!f_v)){
      if(fd_sgn){
        lo = 0x80000000L;
      }
      else{
        lo = 0x00000000L;
      }
    }
  }
  *fdh = (fp_fmt==fmtW)? (lo&0x80000000)? 0xffffffff: 0 : hi;
  *fdl = lo;
}

static void SWAP(int *fts, ulong *ftmh, ulong *ftml,
                 int *fss, ulong *fsmh, ulong *fsml)
{
  int   i;
  ulong u;

  i = *fts;
  *fts = *fss;
  *fss = i;

  u = *ftmh;
  *ftmh = *fsmh;
  *fsmh = u;

  u = *ftml;
  *ftml = *fsml;
  *fsml = u;
}

static void ROUNDL(int rm)
{
  int   sftcnt;
  ulong adj;
  ulong upper;
  ulong tmph,tmpl;
  ulong dmy2,dmy1,dmy0;

  to_fix = 1;
  fd_sgn = fs_sgn;
  if(fp_fmt==fmtD){
    cvt_e_chk = 1;
    sftcnt = fs_exp - 0x3ff - 52;
    if(sftcnt&0x1000){
      sftcnt = ~sftcnt + 1;
      RSHIFTS(fs_manh, fs_manl, sftcnt, &tmph, &tmpl);
      upper = 0;
    }
    else if(sftcnt==0){
      upper = 0;
      tmph = fs_manh;
      tmpl = fs_manl;
    }
    else if(sftcnt<11){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else if((sftcnt==11)&&fs_sgn&&((fs_manh&0x7fffff)==0)&&(fs_manl==0)){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else{ /* overflow */
      cvt_ov = 1;
      upper = 0x03;
      tmph  = 0xffffffff;
      tmpl  = 0xfffffff8;
    }
  }
  else if(fp_fmt==fmtS){
    cvt_e_chk = 1;
    sftcnt = fs_exp - 0x7f - 23;
    if(sftcnt&0x1000){
      sftcnt = ~sftcnt + 1;
      RSHIFTS(fs_manh, fs_manl, sftcnt, &tmph, &tmpl);
      upper = 0;
    }
    else if(sftcnt==0){
      upper = 0;
      tmph  = fs_manh;
      tmpl  = fs_manl;
    }
    else if(sftcnt<40){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else if((sftcnt==40)&&fs_sgn&&((fs_manl&0x03ffffff)==0)){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else{ /* overflow */
      cvt_ov = 1;
      upper = 0x03;
      tmph  = 0xffffffff;
      tmpl  = 0xfffffff8;
    }
  }
  inexact = ((tmpl&7L)!=0);
  switch(rm){
  case rmRN:
    if((tmpl&4L)==0L)  /* Guard Bit */
      adj = 0L;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8L;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8;
    else if(tmpl&8L)  /* Mantissa LSB */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRZ:
    adj = 0L;
    break;
  case rmRP:
    if(fd_sgn)
      adj = 0L;
    else if(tmpl&4L)  /* Guard Bit */
      adj = 8L;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8L;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRM:
    if(fd_sgn==0)
      adj = 0L;
    else if(tmpl&4L)  /* Guard Bit */
      adj = 8;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  }
  IADD96(upper, tmph, tmpl, 0L, 0L, adj,
          &dmy2, &dmy1, &dmy0);
  if(cvt_ov||((!fs_sgn)&&(dmy2&0x4))||((fs_sgn)&&(dmy2&0x8))){ /* overflow */
    cvt_ov = 1;
#if 0
    if(fd_sgn){
      fd_manh = 0x80000000L;
      fd_manl = 0x00000000L;
    }
    else{
      fd_manh = 0x7fffffffL;
      fd_manl = 0xffffffffL;
    }
#else
    fd_manh = 0x7fffffffL;
    fd_manl = 0xffffffffL;
#endif
  }
  else{
    tmpl = (dmy0>>3)|(dmy1<<(32-3));
    tmph = (dmy1>>3)|(dmy2<<(32-3));
    if(fd_sgn)
      INEG64(tmph, tmpl, &fd_manh, &fd_manl);
    else{
      fd_manh = tmph;
      fd_manl = tmpl;
    }
  }
  fp_fmt = fmtL;
  round_inh = 1;
  shift_inh = 1;
}

static void ROUNDW(int rm)
{
  int   sftcnt;
  ulong adj;
  ulong upper;
  ulong tmph,tmpl;
  ulong dmy2,dmy1,dmy0;

  to_fix = 1;
  fd_sgn = fs_sgn;
  if(fp_fmt==fmtD){
    cvt_e_chk = 1;
    sftcnt = fs_exp - 0x3ff - 52;
    if(sftcnt< -21){
      sftcnt = ~sftcnt + 1;
      RSHIFTS(fs_manh, fs_manl, sftcnt, &tmph, &tmpl);
      upper = 0;
    }
    else if((sftcnt== -21)&&fs_sgn
          &&((fs_manh&0x007fffff)==0)
          &&((fs_manl&0xff000000)==0)){
      sftcnt = ~sftcnt + 1;
      RSHIFTS(fs_manh, fs_manl, sftcnt, &tmph, &tmpl);
      upper = 0;
    }
    else{ /* overflow */
      cvt_ov = 1;
      upper = 0;
      tmph  = 0x03;
      tmpl  = 0xfffffff8;
    }
  }
  else if(fp_fmt==fmtS){
    cvt_e_chk = 1;
    sftcnt = fs_exp - 0x7f - 23;
    if(sftcnt&0x1000){
      sftcnt = ~sftcnt + 1;
      RSHIFTS(fs_manh, fs_manl, sftcnt, &tmph, &tmpl);
      upper = 0;
    }
    else if(sftcnt<8){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else if((sftcnt==8)&&fs_sgn&&((fs_manl&0x03ffffff)==0)){
      LSHIFT(fs_manh, fs_manl, sftcnt, &upper, &tmph, &tmpl);
    }
    else{ /* overflow */
      cvt_ov = 1;
      upper = 0x00;
      tmph  = 0x03;
      tmpl  = 0xfffffff8;
    }
  }
  inexact = ((tmpl&7L)!=0);
  switch(rm){
  case rmRN:
    if((tmpl&4L)==0L)  /* Guard Bit */
      adj = 0L;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8L;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8;
    else if(tmpl&8L)  /* Mantissa LSB */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRZ:
    adj = 0L;
    break;
  case rmRP:
    if(fd_sgn)
      adj = 0L;
    else if(tmpl&4L)  /* Guard Bit */
      adj = 8L;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8L;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  case rmRM:
    if(fd_sgn==0)
      adj = 0L;
    else if(tmpl&4L)  /* Guard Bit */
      adj = 8L;
    else if(tmpl&2L)  /* Round Bit */
      adj = 8L;
    else if(tmpl&1L)  /* Sticky Bit */
      adj = 8L;
    else
      adj = 0L;
    break;
  }
  IADD96(upper, tmph, tmpl, 0L, 0L, adj,
          &dmy2, &dmy1, &dmy0);
  if(cvt_ov||((!fs_sgn)&&(dmy1&0x4))||((fs_sgn)&&(dmy1&0x8))){ /* overflow */
    cvt_ov = 1;
#if 0
    if(fd_sgn){
      fd_manh = 0x00000000L;
      fd_manl = 0x80000000L;
    }
    else{
      fd_manh = 0x00000000;
      fd_manl = 0x7fffffff;
    }
#else
    fd_manh = 0x00000000;
    fd_manl = 0x7fffffff;
#endif
  }
  else{
    tmpl = (dmy0>>3)|(dmy1<<(32-3));
    tmph = (dmy1>>3)|(dmy2<<(32-3));
    if(fd_sgn)
      INEG64(tmph, tmpl, &fd_manh, &fd_manl);
    else{
      fd_manh = tmph;
      fd_manl = tmpl;
    }
  }
  fp_fmt = fmtW;
  round_inh = 1;
  shift_inh = 1;
}

static void FP_ADD(void)
{
  int   del_exp;
  int   dummy;
  ulong fs_man0h, fs_man0l;
  ulong ft_man0h, ft_man0l;
  ulong fs_man1h, fs_man1l;
  ulong ft_man1h, ft_man1l;
  ulong fd_man0h, fd_man0l;
  ulong fd_man1h, fd_man1l;

  del_exp   = fs_exp - ft_exp;
  if(del_exp&0x800){ /* fs_exp < ft_exp */
    SWAP(&ft_sgn, &ft_manh, &ft_manl,
         &fs_sgn, &fs_manh, &fs_manl);
    del_exp = -del_exp;
    fd_exp = ft_exp;
  }
  else{
    fd_exp = fs_exp;
  }
  fs_man0h = fs_manh;
  fs_man0l = fs_manl;
  RSHIFTS(ft_manh, ft_manl, del_exp, &ft_man0h, &ft_man0l);
  if(fs_sgn){
    INEG64(fs_man0h, fs_man0l, &fs_man1h, &fs_man1l);
  }
  else{
    fs_man1h = fs_man0h;
    fs_man1l = fs_man0l;
  }
  if(ft_sgn){
    INEG64(ft_man0h, ft_man0l, &ft_man1h, &ft_man1l);
  }
  else{
    ft_man1h = ft_man0h;
    ft_man1l = ft_man0l;
  }
  IADD64(fs_man1h, fs_man1l, ft_man1h, ft_man1l, &fd_man0h, &fd_man0l, &dummy);
  fd_sgn = ((fd_man0h&0x80000000L)!=0L);
  if(fd_sgn){
    INEG64(fd_man0h, fd_man0l, &fd_man1h, &fd_man1l);
  }
  else{
    fd_man1h = fd_man0h;
    fd_man1l = fd_man0l;
  }
  NORMALIZE(fd_man1h, fd_man1l);
}

static void FP_SUB(void)
{
  ft_sgn ^= 1;
  FP_ADD();
}

static void FP_MUL(void)
{
  ulong tmp3,tmp2,tmp1,tmp0;
  int   over;

  if(fp_fmt==fmtD){
    fd_sgn = fs_sgn ^ ft_sgn;
    fd_exp = fs_exp + ft_exp - 0x3ff;
    UMUL64(fs_manh, fs_manl, ft_manh, ft_manl,
           &tmp3,   &tmp2,   &tmp1,   &tmp0);
    fd_manh = (tmp3<<9)|(tmp2>>23);
    fd_manl = (tmp2<<9)|(tmp1>>23)|((tmp1&0x007fffffL)!=0)|(tmp0!=0L);
    over    = ((fd_manh&0x01000000L)!=0L);
  }
  else{
    fd_sgn = fs_sgn ^ ft_sgn;
    fd_exp = fs_exp + ft_exp - 0x7f;
    UMUL32(fs_manl, ft_manl, &tmp1, &tmp0);
    fd_manh = 0L;
    fd_manl = (tmp1<<6)|(tmp0>>26)|((tmp0&0x03ffffffL)!=0L);
    over    = ((fd_manl&0x08000000L)!=0L);
  }
  if((fd_manh|fd_manl)==0L){
    fd_exp = 0;
  }
  if(over){
    fd_exp++;
    RSHIFTS(fd_manh, fd_manl, 1, &tmp1, &tmp0);
    fd_manh = tmp1;
    fd_manl = tmp0;
  }
}

static void FP_DIV(void)
{
  ulong qh, ql, rh, rl;

  if(fp_fmt==fmtD){
    fd_sgn = fs_sgn ^ ft_sgn;
    fd_exp = fs_exp - ft_exp + 0x3ff;
    FDIV64(fs_manh, fs_manl, ft_manh, ft_manl, &qh, &ql, &rh, &rl, 56);
    if(qh&(1<<(56-32))){
      fd_manh = qh>>1;
      fd_manl = (qh<<31)|(ql>>1)|((ql&1L)!=0L)|((rh|rl)!=0L);
    }
    else{
      fd_exp--;
      fd_manh = qh;
      fd_manl = ql|((rh|rl)!=0L);
    }
  }
  else{
    fd_sgn = fs_sgn ^ ft_sgn;
    fd_exp = fs_exp - ft_exp + 0x7f;
    FDIV64(fs_manh, fs_manl, ft_manh, ft_manl, &qh, &ql, &rh, &rl, 27);
    if(ql&(1<<27)){
      fd_manh = 0L;
      fd_manl = (ql>>1)|((ql&1L)!=0L)|((rh|rl)!=0L);
    }
    else{
      fd_exp--;
      fd_manh = 0L;
      fd_manl = ql|((rh|rl)!=0L);
    }
  }
  if((fd_manh|fd_manl)==0L){
    fd_exp = 0;
  }
}

static void FP_SQRT(void)
{
  ulong qh, ql, rl;

  if(fp_fmt==fmtD){
    fd_sgn = fs_sgn;
    if((fs_exp-0x3ff)&1){
      fs_manh = (fs_manh<<1)|(fs_manl>>31);
      fs_manl = fs_manl<<1;
    }
    fd_exp = ((fs_exp-0x3ff)>>1)+0x3ff;
    FSQRT64(fs_manh, fs_manl, &qh, &ql, &rl);
    if(qh&(1<<(56-32))){
      fd_manh = qh>>1;
      fd_manl = (qh<<31)|(ql>>1)|((ql&1L)!=0L)|(rl!=0L);
    }
    else{
      fd_exp--;
      fd_manh = qh;
      fd_manl = ql|(rl!=0L);
    }
  }
  else{
    fd_sgn = fs_sgn;
    if((fs_exp-0x7f)&1){
      fs_manl = fs_manl<<1;
    }
    fd_exp = ((fs_exp-0x7f)>>1)+0x7f;
    FSQRT32(fs_manl, &ql, &rl);
    if(ql&(1<<27)){
      fd_manh = 0L;
      fd_manl = (ql>>1)|((ql&1L)!=0L)|(rl!=0L);
    }
    else{
      fd_exp--;
      fd_manh = 0L;
      fd_manl = ql|(rl!=0L);
    }
  }
  if((fd_manh|fd_manl)==0L){
    fd_exp = 0;
  }
}

static void FP_ABS(void)
{
  fd_sgn = 0;
  fd_exp = fs_exp;
  fd_manh = fs_manh;
  fd_manl = fs_manl;
}

static void FP_NEG(void)
{
  fd_sgn = fs_sgn^1;
  fd_exp = fs_exp;
  fd_manh = fs_manh;
  fd_manl = fs_manl;
}

static void FP_RECIP(void)
{
  ft_sgn = fs_sgn;
  ft_exp = fs_exp;
  ft_manh = fs_manh;
  ft_manl = fs_manl;
  if(fp_fmt==fmtD){
    fs_sgn = 0;
    fs_exp = 0x3ff;
    fs_manh = 0x00800000L;
    fs_manl = 0x00000000L;
  }
  else{
    fs_sgn = 0;
    fs_exp = 0x7f;
    fs_manh = 0x00000000L;
    fs_manl = 0x04000000L;
  }
  FP_DIV();
}

static void FP_RSQRT(void)
{
  FP_SQRT();
  fs_sgn = fd_sgn;
  fs_exp = fd_exp;
  fs_manh = fd_manh;
  fs_manl = fd_manl;
  FP_RECIP();
}

static void FP_CVTD(void)
{
  ulong up;
  int   clz;

  if(fp_fmt==fmtS){
    if(fs_infi){
      fd_sgn = fs_sgn;
      fd_exp = 0x7ff;
      fd_manh = 0;
      fd_manl = 0;
    }
    else{
      cvt_e_chk = 1;
      fd_sgn = fs_sgn;
      fd_exp = (fs_exp)? (fs_exp - 0x7f +0x3ff) : 0;
      LSHIFT(fs_manh, fs_manl, 29, &up, &fd_manh, &fd_manl);
    }
  }
  else if(fp_fmt==fmtL){
    clz = CLZ64(fs_manh, fs_manl);
    if(clz==0){ /* -1 */
      fd_sgn = fs_sgn;
      fd_exp = 0x3ff+63;
      fd_manh = (1<<(55-32));
      fd_manl = 0;
    }
    else if(clz<8){
      fd_sgn = fs_sgn;
      fd_exp = (0x3ff + (63-clz)) ;
      RSHIFTS(fs_manh, fs_manl, 8-clz, &fd_manh, &fd_manl);
    }
    else if(clz==8){
      fd_sgn = fs_sgn;
      fd_exp = 0x3ff + 55;
      fd_manh = fs_manh;
      fd_manl = fs_manl;
    }
    else if(clz==64){ /* 0 */
      fd_sgn = fs_sgn;
      fd_exp = 0;
      fd_manh = 0L;
      fd_manl = 0L;
    }
    else{
      fd_sgn = fs_sgn;
      fd_exp = (fs_manh|fs_manl)? (0x3ff - (clz-63)) : 0;
      LSHIFT(fs_manh, fs_manl, clz-8, &up, &fd_manh, &fd_manl);
    }
  }
  else if(fp_fmt==fmtW){
    clz = CLZ32(fs_manl);
    if(clz==32){ /* 0 */
      fd_sgn = fs_sgn;
      fd_exp = 0;
      fd_manh = 0L;
      fd_manl = 0L;
    }
    else{
      fd_sgn = fs_sgn;
      fd_exp = (fs_manl)? (0x3ff + (31-clz)) : 0;
      LSHIFT(fs_manh, fs_manl, clz+24, &up, &fd_manh, &fd_manl);
    }
  }
  fp_fmt = fmtD;
}

static void FP_CVTS(void)
{
  ulong up;
  int   clz;

  if(fp_fmt==fmtD){
    if(fs_infi){
      fd_sgn = fs_sgn;
      fd_exp = 0xff;
      fd_manh = 0;
      fd_manl = 0;
    }
    else{
      cvt_e_chk = 1;
      fd_sgn = fs_sgn;
      fd_exp = (fs_exp)? (fs_exp - 0x3ff +0x7f) : 0;
      RSHIFTS(fs_manh, fs_manl, 29, &fd_manh, &fd_manl);
    }
  }
  else if(fp_fmt==fmtL){
    clz = CLZ64(fs_manh, fs_manl);
    if(clz==0){ /* -1 */
      fd_sgn = fs_sgn;
      fd_exp = 0x7f+31+32;
      fd_manh = 0;
      fd_manl = (1<<26);
    }
    else if(clz<37){
      fd_sgn = fs_sgn;
      fd_exp = (0x7f + (63-clz)) ;
      RSHIFTS(fs_manh, fs_manl, 37-clz, &fd_manh, &fd_manl);
    }
    else if(clz==37){
      fd_sgn = fs_sgn;
      fd_exp = 0x7f + 26;
      fd_manh = fs_manh;
      fd_manl = fs_manl;
    }
    else if(clz==64){ /* 0 */
      fd_sgn = fs_sgn;
      fd_exp = 0;
      fd_manh = 0L;
      fd_manl = 0L;
    }
    else{
      fd_sgn = fs_sgn;
      fd_exp = (fs_manh|fs_manl)? (0x7f - (clz-63)) : 0;
      LSHIFT(fs_manh, fs_manl, clz-37, &up, &fd_manh, &fd_manl);
    }
  }
  else if(fp_fmt==fmtW){
    clz = CLZ32(fs_manl);
    if(clz<5){
      fd_sgn = fs_sgn;
      fd_exp = (fs_manl)? (0x7f + (31-clz)) : 0;
      RSHIFTS(fs_manh, fs_manl, 5-clz, &fd_manh, &fd_manl);
    }
    else if(clz==32){ /* 0 */
      fd_sgn = fs_sgn;
      fd_exp = 0;
      fd_manh = 0L;
      fd_manl = 0L;
    }
    else{
      fd_sgn = fs_sgn;
      fd_exp = (fs_manl)? (0x7f + (31-clz)) : 0;
      LSHIFT(fs_manh, fs_manl, clz-5, &up, &fd_manh, &fd_manl);
    }
  }
  fp_fmt = fmtS;
}

static void FP_CMP(ulong sh, ulong sl, ulong th, ulong tl, int cond, int *cc)
{
  int less,equal,unorder;
  int cond3,cond2,cond1,cond0;

  cond3 = ((cond&8)!=0);
  cond2 = ((cond&4)!=0);
  cond1 = ((cond&2)!=0);
  cond0 = ((cond&1)!=0);

  if(fs_qnan||ft_qnan||fs_snan||ft_snan){
    less    = 0;
    equal   = 0;
    unorder = 1;
    if(cond3||fs_snan||ft_snan)
      f_v = 1;
  }
  else{
    if(fp_fmt==fmtS){
      ACMP64(sign_extend(sl), sl, sign_extend(tl), tl, &less, &equal);
      if(((sl|tl)&0x7fffffffL)==0L){
	less = 0;
	equal = 1;
      }
    }
    else{
      ACMP64(sh, sl, th, tl, &less, &equal);
      if((((sh|th)&0x7fffffffL)==0L)&&((sl|tl)==0L)){
	less = 0;
	equal = 1;
      }
    }
    unorder = 0;
  }
#if 0
  *cc = (!f_v)&&((cond2 & less)|(cond1 & equal)|(cond0 & unorder));
#else
  *cc = ((cond2 & less)|(cond1 & equal)|(cond0 & unorder));
#endif
}

static void FP_MADD(void)
{
  int x_sgn;
  fd_infi = fs_infi||ft_infi;
  x_sgn   = fs_sgn ^ ft_sgn;
  operand1 = 0;
  FP_MUL();
  fs_sgn  = fd_sgn;
  fs_exp  = fd_exp;
  fs_manh = fd_manh;
  fs_manl = fd_manl;
  ft_sgn  = fr_sgn;
  ft_exp  = fr_exp;
  ft_manh = fr_manh;
  ft_manl = fr_manl;
  if((!fr_infi)
#ifdef WILDCAT
     &&(fp_fs==0)
#endif
     &&((fd_exp<0)||((fd_exp==0)&&((fd_manh|fd_manl)!=0)))){
    madd_e = 1;
  }
  else if((!fr_infi)&&(!fd_infi)
        &&(((fp_fmt==fmtS)&&(fd_exp>=0xff))||((fp_fmt==fmtD)&&(fd_exp>=0x7ff)))){
    madd_e = 1;
  }
  else{
#ifdef WILDCAT
    if((fp_fs==1)&&(fd_exp<=0)){
      fs_exp = 0;
/*
      fs_manh = 0;
      fs_manl = 0;
*/
    }
#endif
    FP_ADD();
    if(fd_infi){
      fd_sgn = x_sgn;
      fs_infi = fd_infi;
      fs_sgn  = fd_sgn;
      ft_infi = fr_infi;
      ft_sgn  = fr_sgn;
    }
    if(fd_infi&&fr_infi){
      fd_infi = (fd_sgn==ft_sgn);
    }
    else{
      fd_infi = (fd_infi||fr_infi);
      if(fr_infi){
	fd_sgn = fr_sgn;
      }
    }
  }
}

static void FP_MSUB(void)
{
  int x_sgn;
  fd_infi = fs_infi||ft_infi;
  x_sgn   = fs_sgn ^ ft_sgn;
  operand1 = 0;
  FP_MUL();
  fs_sgn  = fd_sgn;
  fs_exp  = fd_exp;
  fs_manh = fd_manh;
  fs_manl = fd_manl;
  ft_sgn  = fr_sgn;
  ft_exp  = fr_exp;
  ft_manh = fr_manh;
  ft_manl = fr_manl;
  if((!fr_infi)
#ifdef WILDCAT
     &&(fp_fs==0)
#endif
     &&((fd_exp<0)||((fd_exp==0)&&((fd_manh|fd_manl)!=0)))){
    madd_e = 1;
  }
  else if((!fr_infi)&&(!fd_infi)
        &&(((fp_fmt==fmtS)&&(fd_exp>=0xff))||((fp_fmt==fmtD)&&(fd_exp>=0x7ff)))){
    madd_e = 1;
  }
  else{
#ifdef WILDCAT
    if((fp_fs==1)&&(fd_exp<=0)){
      fs_exp = 0;
/*
      fs_manh = 0;
      fs_manl = 0;
*/
    }
#endif
    FP_SUB();
    if(fd_infi){
      fd_sgn = x_sgn;
      fs_infi = fd_infi;
      fs_sgn  = fd_sgn;
      ft_infi = fr_infi;
      ft_sgn  = fr_sgn^1;
    }
    if(fd_infi&&fr_infi){
      fd_infi = (fd_sgn==ft_sgn);
    }
    else{
      fd_infi = (fd_infi||fr_infi);
      if(fr_infi){
	fd_sgn = fr_sgn^1;
      }
    }
  }
}

static void FP_NMADD(void)
{
  int x_sgn;
  fd_infi = fs_infi||ft_infi;
  x_sgn   = fs_sgn ^ ft_sgn;
  operand1 = 0;
  FP_MUL();
  fs_sgn  = fd_sgn;
  fs_exp  = fd_exp;
  fs_manh = fd_manh;
  fs_manl = fd_manl;
  ft_sgn  = fr_sgn;
  ft_exp  = fr_exp;
  ft_manh = fr_manh;
  ft_manl = fr_manl;
  if((!fr_infi)
#ifdef WILDCAT
     &&(fp_fs==0)
#endif
     &&((fd_exp<0)||((fd_exp==0)&&((fd_manh|fd_manl)!=0)))){
    madd_e = 1;
  }
  else if((!fr_infi)&&(!fd_infi)
       && (((fp_fmt==fmtS)&&(fd_exp>=0xff))||((fp_fmt==fmtD)&&(fd_exp>=0x7ff)))){
    madd_e = 1;
  }
  else{
#ifdef WILDCAT
    if((fp_fs==1)&&(fd_exp<=0)){
      fs_exp = 0;
/*
      fs_manh = 0;
      fs_manl = 0;
*/
    }
#endif
    FP_ADD();
    if(fd_infi){
      fd_sgn = x_sgn;
      fs_infi = fd_infi;
      fs_sgn  = fd_sgn;
      ft_infi = fr_infi;
      ft_sgn  = fr_sgn;
    }
    if(fd_infi&&fr_infi){
      fd_infi = (fd_sgn==ft_sgn);
    }
    else{
      fd_infi = (fd_infi||fr_infi);
      if(fr_infi){
	fd_sgn = fr_sgn;
      }
    }
  }
  fd_sgn ^= 1;
}

static void FP_NMSUB(void)
{
  int x_sgn;
  fd_infi = fs_infi||ft_infi;
  x_sgn   = fs_sgn ^ ft_sgn;
  operand1 = 0;
  FP_MUL();
  fs_sgn  = fd_sgn;
  fs_exp  = fd_exp;
  fs_manh = fd_manh;
  fs_manl = fd_manl;
  ft_sgn  = fr_sgn;
  ft_exp  = fr_exp;
  ft_manh = fr_manh;
  ft_manl = fr_manl;
  if((!fr_infi)
#ifdef WILDCAT
     &&(fp_fs==0)
#endif
     &&((fd_exp<0)||((fd_exp==0)&&((fd_manh|fd_manl)!=0)))){
    madd_e = 1;
  }
  else if((!fr_infi)&&(!fd_infi)
        &&(((fp_fmt==fmtS)&&(fd_exp>=0xff))||((fp_fmt==fmtD)&&(fd_exp>=0x7ff)))){
    madd_e = 1;
  }
  else{
#ifdef WILDCAT
    if((fp_fs==1)&&(fd_exp<=0)){
      fs_exp = 0;
/*
      fs_manh = 0;
      fs_manl = 0;
*/
    }
#endif
    FP_SUB();
    if(fd_infi){
      fd_sgn = x_sgn;
      fs_infi = fd_infi;
      fs_sgn  = fd_sgn;
      ft_infi = fr_infi;
      ft_sgn  = fr_sgn^1;
    }
    if(fd_infi&&fr_infi){
      fd_infi = (fd_sgn==ft_sgn);
    }
    else{
      fd_infi = (fd_infi||fr_infi);
      if(fr_infi){
	fd_sgn = fr_sgn^1;
      }
    }
  }
  fd_sgn ^= 1;
}

static void V_CHECK(void)
{
  int case1, case2, case3, case4, case5, case6;
  int op_add, op_sub, op_div, op_sqrt;
  int xsgn, esgn;

  op_add = ((fp_op==opADD)||(fp_op==opMADD_S)||(fp_op==opMADD_D)||(fp_op==opNMADD_S)||(fp_op==opNMADD_D));
  op_sub = ((fp_op==opSUB)||(fp_op==opMSUB_S)||(fp_op==opMSUB_D)||(fp_op==opNMSUB_S)||(fp_op==opNMSUB_D));
  op_div = (fp_op==opDIV);
  op_sqrt = ((fp_op==opSQRT)||(fp_op==opRSQRT));
  xsgn   = fs_sgn^ft_sgn;
  esgn   = (xsgn==0);

  case1 = (op_add&&(!fp_cmp))? (xsgn&&fs_infi&&ft_infi) : 0;
  case2 = (op_sub&&(!fp_cmp))? (xsgn&&fs_infi&&ft_infi) : 0;
  case3 = mul_err;
  case4 = (op_div)? ((fs_zero&&ft_zero)||(fs_infi&&ft_infi)) : 0;
  case5 = (operand1)? fs_snan : (fs_snan||ft_snan||(madd_op&&fr_snan));
  case6 = sqr_err;
  f_v   |= (case1||case2||case3||case4||case5||case6||cvt_ov);
}

void O_CHECK(void)
{
  int exp_max;
  int is_over;

  exp_max = (fp_fmt==fmtD)? 0x7fe : 0xfe;
  is_over = fd_exp - exp_max;
  f_o = ((is_over>0)&&(!fd_infi)&&(!to_fix));
}

void U_CHECK(void)
{
  int zero;
  zero = ((fd_manh|fd_manl)==0);
  f_u = ((fd_exp<0)||((!zero)&&(fd_exp==0)))&&(!to_fix)&&(!fd_zero)&&(!fd_infi);
}

void E_CHECK(void)
{
  int deno_src,deno_src2;
  int qnan_src,qnan_src2;
  int snan_src;

  deno_src = (operand1)? fs_deno : (fs_deno||ft_deno||(madd_op&&fr_deno));
  qnan_src = (operand1)? fs_qnan : (fs_qnan||ft_qnan||(madd_op&&fr_qnan));
  snan_src = (operand1)? fs_snan : (fs_snan||ft_snan||(madd_op&&fr_snan));

  deno_src2 = ((!fp_cmp)&&deno_src&&(cvt_e_chk||(fp_fmt==fmtS)||(fp_fmt==fmtD)));
  qnan_src2 = ((!fp_cmp)&&qnan_src);

  if(deno_src2||qnan_src2){
    f_e = 1;
  }
  else if(snan_src){
    f_e = 0;
    f_v = 1;
  }
  else{
    f_e = ((!fp_cmp)&&(!fp_fs)&&f_u);
  }
}

static void EXCEPTION(int *e, int *v, int *z, int *o, int *u, int *i)
{
  int org_inf;

  V_CHECK();
  O_CHECK();
  U_CHECK();
  E_CHECK();
  org_inf = fs_infi||((!operand1)&&ft_infi)||(madd_op&&fr_infi);
  *e = f_e;
  *v = (!f_e)&&f_v;
  *z = (!f_e)&&(!(*v))&&f_z&&(!fp_cmp);
  *o = (!f_e)&&(!(*v))&&(!(*z))&&f_o&&(!fp_cmp)&&(!org_inf);
  *u = (!f_e)&&(!(*v))&&(!(*z))&&f_u&&(!fp_cmp);
  *i = ((*o)||(*u)||(inexact&&(!org_inf)))&&(!(*v))&&(!(*z))&&(!fp_cmp);
  fp_except = (*e)||(*v)||(*z)||(*o)||(*u);
}

static void SP_RES(ulong *hi, ulong *lo)
{
  ulong qnanh, qnanl;
  ulong infih, infil;
  ulong maxh, maxl;
  ulong minh, minl;
  ulong zeroh, zerol;
  ulong sgnh, sgnl;

  if(fp_fmt==fmtD){
    qnanh = 0x7ff7ffffL;
    qnanl = 0xffffffffL;
    infih = 0x7ff00000L;
    infil = 0x00000000L;
    maxh  = 0x7fefffffL;
    maxl  = 0xffffffffL;
    minh  = 0x00100000L;
    minl  = 0x00000000L;
    zeroh = 0x80000000L;
    zerol = 0x00000000L;
  }
  else{
    qnanh = 0x00000000L;
    qnanl = 0x7fbfffffL;
    infih = 0x00000000L;
    infil = 0x7f800000L;
    maxh  = 0x00000000L;
    maxl  = 0x7f7fffffL;
    minh  = 0x00000000L;
    minl  = 0x00800000L;
    zeroh = 0x00000000L;
    zerol = 0x80000000L;
  }
  if(fd_sgn){
    sgnh = zeroh;
    sgnl = zerol;
  }
  else{
    sgnh = 0L;
    sgnl = 0L;
  }
  if(f_v){
    if(cvt_ov){
      *hi = fd_manh;
      *lo = fd_manl;
    }
    else{
      *hi = qnanh;
      *lo = qnanl;
    }
  }
  else if(f_z){
    *hi = infih|sgnh;
    *lo = infil|sgnl;
  }
  else if(f_o){
    switch(fp_rm){
    case rmRN:
      *hi = infih|sgnh;
      *lo = infil|sgnl;
      break;
    case rmRZ:
      *hi = maxh|sgnh;
      *lo = maxl|sgnl;
      break;
    case rmRP:
      if(fd_sgn){
        *hi = maxh|sgnh;
        *lo = maxl|sgnl;
      }
      else{
        *hi = infih;
        *lo = infil;
      }
      break;
    case rmRM:
      if(fd_sgn){
        *hi = infih|sgnh;
        *lo = infil|sgnl;
      }
      else{
        *hi = maxh;
        *lo = maxl;
      }
      break;
    }
  }
  else{
    switch(fp_rm){
    case rmRN:
    case rmRZ:
      *hi = sgnh;
      *lo = sgnl;
      break;
    case rmRP:
      if(fd_sgn){
        *hi = zeroh;
        *lo = zerol;
      }
      else{
        *hi = minh;
        *lo = minl;
      }
      break;
    case rmRM:
      if(fd_sgn){
        *hi = minh|sgnh;
        *lo = minl|sgnl;
      }
      else{
        *hi = 0L;
        *lo = 0L;
      }
      break;
    }
  }
}

void FPU(int   op,               /* operation */
         int   fmt,              /* format    */
	 int   rm,               /* rounding mode */
         int   fs_bit,           /* FS bit    */
         ulong fsh,  ulong fsl,  /* source    */
         ulong fth,  ulong ftl,  /* target    */
         ulong *fdh, ulong *fdl, /* destination */
         int *e,   /* Unimplement Operation */
         int *v,   /* Invalid Operation     */
         int *z,   /* Divide by Zero        */
         int *o,   /* Overflow              */
         int *u,   /* Underflow             */
         int *i,   /* Inexact Operation     */
         int *cmp, /* Compare?              */
         int *cc   /* Condition Code        */
)
{
    fp_op  = op;
	fp_rm  = rm;
    fp_fmt = fmt;
    fp_cmp = 0;
    fp_fs  = fs_bit;

    madd_op = 0;
    round_inh = 0;
    shift_inh = 0;

    cvt_e_chk = 0;
    cvt_ov = 0;
    f_e    = 0;
    f_v    = 0;
    f_z    = 0;
    f_o    = 0;
    f_u    = 0;
    f_i    = 0;
    f_cc   = 0;

    UNPACK(fsh, fsl, &fs_sgn, &fs_exp, &fs_manh, &fs_manl,
           &fs_qnan, &fs_snan, &fs_zero, &fs_deno, &fs_infi);
    UNPACK(fth, ftl, &ft_sgn, &ft_exp, &ft_manh, &ft_manl,
           &ft_qnan, &ft_snan, &ft_zero, &ft_deno, &ft_infi);

    abs_neg = 0;
    fr_qnan = 0;
    fr_snan = 0;
    fr_zero = 0;
    fr_deno = 0;
    fr_infi = 0;
    to_fix  = 0;
    mul_err = 0;
    sqr_err = 0;
    operand1 = 1;
    fd_infi = 0;
    fd_zero = 0;

    switch(fp_op){
    case opADD:
      if((fmt==fmtS)||(fmt==fmtD)){
        fd_infi = (fs_infi&&(!ft_infi))||
                  ((!fs_infi)&&ft_infi)||
                  ((fs_sgn==ft_sgn)&&fs_infi&&ft_infi);
        operand1 = 0;
        FP_ADD();
      }
      else
        f_e = 1;
      break;
    case opSUB:
      if((fmt==fmtS)||(fmt==fmtD)){
        fd_infi = (fs_infi&&(!ft_infi))||
                  ((!fs_infi)&&ft_infi)||
                  ((fs_sgn!=ft_sgn)&&fs_infi&&ft_infi);
        operand1 = 0;
        FP_SUB();
      }
      else
        f_e = 1;
      break;
    case opMUL:
      if((fmt==fmtS)||(fmt==fmtD)){
        fd_infi = fs_infi||ft_infi;
        operand1 = 0;
        mul_err = ((fs_infi&&ft_zero)||(fs_zero&&ft_infi));
        FP_MUL();
      }
      else
        f_e = 1;
      break;
    case opDIV:
      if((fmt==fmtS)||(fmt==fmtD)){
        fd_infi = ((!fs_zero)&&ft_zero)||
                  (fs_infi&&(!ft_infi));
        fd_zero = ft_infi;
        operand1 = 0;
        if((!fs_zero)&&(!fs_infi)&&ft_zero)
          f_z = 1;
        FP_DIV();
      }
      else
        f_e = 1;
      break;
    case opSQRT:
      if((fmt==fmtS)||(fmt==fmtD)){
        fd_infi = fs_infi;
        sqr_err = (fs_sgn&&(!fs_zero));
        FP_SQRT();
      }
      else
        f_e = 1;
      break;
    case opABS:
      if((fmt==fmtS)||(fmt==fmtD)){
	abs_neg = 1;
        FP_ABS();
      }
      else
        f_e = 1;
      break;
    case opNEG:
      if((fmt==fmtS)||(fmt==fmtD)){
	abs_neg = 1;
        FP_NEG();
      }
      else
        f_e = 1;
      break;
    case opROUND_L:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDL(rmRN);
      else
        f_e = 1;
      break;
    case opTRUNC_L:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDL(rmRZ);
      else
        f_e = 1;
      break;
    case opCEIL_L:
       if((fmt==fmtS)||(fmt==fmtD))
         ROUNDL(rmRP);
      else
        f_e = 1;
      break;
    case opFLOOR_L:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDL(rmRM);
      else
        f_e = 1;
      break;
    case opROUND_W:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDW(rmRN);
      else
        f_e = 1;
      break;
    case opTRUNC_W:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDW(rmRZ);
      else
        f_e = 1;
      break;
    case opCEIL_W:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDW(rmRP);
      else
        f_e = 1;
      break;
    case opFLOOR_W:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDW(rmRM);
      else
        f_e = 1;
      break;
    case opCVT_L:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDL(rm);
      else
        f_e = 1;
      break;
    case opCVT_W:
      if((fmt==fmtS)||(fmt==fmtD))
        ROUNDW(rm);
      else
        f_e = 1;
      break;
    case opCVT_D:
      if((fmt==fmtS)||(fmt==fmtL)||(fmt==fmtW)){
	if(fs_infi&&(fmt==fmtS)){
	  fd_infi = 1;
	  fd_sgn  = fs_sgn;
	  fp_fmt = fmtD;
	  abs_neg = 1;
	}
	else
	  FP_CVTD();
      }
      else
        f_e = 1;
      break;
    case opCVT_S:
      if((fmt==fmtD)||(fmt==fmtL)||(fmt==fmtW)){
	if(fs_infi&&(fmt==fmtD)){
	  fd_infi = 1;
	  fd_sgn  = fs_sgn;
	  fp_fmt = fmtS;
	  abs_neg = 1;
	}
	else
	  FP_CVTS();
      }
      else
        f_e = 1;
      break;
    case opRECIP:
      if((fmt==fmtS)||(fmt==fmtD)){
	if(fs_infi){
	  fd_zero = 1;
	  fd_sgn  = fs_sgn;
	  abs_neg = 1;
	}
        else if(fs_zero){
          f_z = 1;
	  fd_infi = 1;
	  fd_sgn = fs_sgn;
	  abs_neg = 1;
        }
	else
	  FP_RECIP();
      }
      else
        f_e = 1;
      break;
    case opRSQRT:
      if((fmt==fmtS)||(fmt==fmtD)){
	if(fs_zero){
          f_z = 1;
	  fd_sgn = fs_sgn;
	  abs_neg = 1;
	}
	else if(fs_sgn){
	  f_v = 1;
	  abs_neg = 1;
	}
	else if(fs_infi){
	  fd_zero = 1;
	  fd_sgn  = fs_sgn;
	  abs_neg = 1;
	}
        else
	  FP_RSQRT();
      }
      else
        f_e = 1;
      break;
    case opCMP_F:
    case opCMP_UN:
    case opCMP_EQ:
    case opCMP_UEQ:
    case opCMP_OLT:
    case opCMP_ULT:
    case opCMP_OLE:
    case opCMP_ULE:
    case opCMP_SF:
    case opCMP_NGLE:
    case opCMP_SEQ:
    case opCMP_NGL:
    case opCMP_LT:
    case opCMP_NGE:
    case opCMP_LE:
    case opCMP_NGT:
      if((fmt==fmtS)||(fmt==fmtD)){
        operand1 = 0;
        fp_cmp = 1;
        FP_CMP(fsh, fsl, fth, ftl, fp_op&0xf, cc);
      }
      else
        f_e = 1;
      break;
    default:
      f_e = 1;
    }
    if(f_e){
      *e = 1;
    }
    else{
      if(abs_neg==0){
	if(round_inh==0)
	  ROUND(rm);
	EXCEPTION(e, v, z, o, u, i);
      }
      else{
	*e = 0;
	*v = f_v;
	*z = f_z;
	*o = 0;
	*u = 0;
	*i = 0;
	fp_except = (f_z||f_v);
	if(fs_deno || fs_qnan){
	  *e = 1;
	  fp_except = 1;
	}
	else if(fs_snan){
	  f_v = 1;
	  *v = 1;
	  fp_except = 1;
	}
      }
      if(!fp_except)
        REPACK(fdh, fdl);
      else
        SP_RES(fdh, fdl);
   }
   if(fp_cmp){
	*z = 0;
	*o = 0;
	*u = 0;
	*i = 0;
   }
  *cmp = fp_cmp;
}

void FPUX(int   op,               /* operation */
	 int   rm,               /* rounding mode */
         int   fs_bit,           /* FS bit    */
         ulong fsh,  ulong fsl,  /* source    */
         ulong fth,  ulong ftl,  /* target    */
	 ulong frh,  ulong frl,  /* rest?     */
         ulong *fdh, ulong *fdl, /* destination */
         int *e,   /* Unimplement Operation */
         int *v,   /* Invalid Operation     */
         int *z,   /* Divide by Zero        */
         int *o,   /* Overflow              */
         int *u,   /* Underflow             */
         int *i   /* Inexact Operation     */
)
{
    int madd_v0,madd_e0;
    fp_op  = op;
	fp_rm  = rm;
    fp_fmt = (op&1)? fmtD: fmtS;
    fp_cmp = 0;
    fp_fs  = fs_bit;
    fd_infi = 0;
    fd_zero = 0;
    operand1 = 0;

    to_fix = 0;
    madd_op = 1;
    madd_e = 0;
    shift_inh = 0;
    cvt_e_chk = 0;
    cvt_ov = 0;
    f_e    = 0;
    f_v    = 0;
    f_z    = 0;
    f_o    = 0;
    f_u    = 0;
    f_i    = 0;
    f_cc   = 0;

    UNPACK(fsh, fsl, &fs_sgn, &fs_exp, &fs_manh, &fs_manl,
           &fs_qnan, &fs_snan, &fs_zero, &fs_deno, &fs_infi);
    UNPACK(fth, ftl, &ft_sgn, &ft_exp, &ft_manh, &ft_manl,
           &ft_qnan, &ft_snan, &ft_zero, &ft_deno, &ft_infi);
    UNPACK(frh, frl, &fr_sgn, &fr_exp, &fr_manh, &fr_manl,
           &fr_qnan, &fr_snan, &fr_zero, &fr_deno, &fr_infi);

    madd_v0 = fs_snan || ft_snan || fr_snan;
    madd_e0 = fs_qnan || ft_qnan || fr_qnan
            ||fs_deno || ft_deno || fr_deno;
    mul_err = (fs_infi&&ft_zero)||(fs_zero&&ft_infi);
    madd_v0 = madd_v0 || mul_err;
    sqr_err = 0;
    madd_e = 0;

    if((!madd_v0)&&(!madd_e0)){
      switch(fp_op){
      case opMADD_S:
      case opMADD_D:
	FP_MADD();
	break;
      case opMSUB_S:
      case opMSUB_D:
	FP_MSUB();
	break;
      case opNMADD_S:
      case opNMADD_D:
	FP_NMADD();
	break;
      case opNMSUB_S:
      case opNMSUB_D:
	FP_NMSUB();
	break;
      default:
	f_e = 1;
      }
      if((!fd_infi)&&((fd_manh|fd_manl)==0))
	fd_sgn = 0;
    }
    if(madd_e0){
      *e = 1;
      *v = 0;
      *z = 0;
      *o = 0;
      *u = 0;
      *i = 0;
    }
    else if(madd_v0){
      *e = 0;
      *v = 1;
      *z = 0;
      *o = 0;
      *u = 0;
      *i = 0;
      if(fp_fmt==fmtS){
	*fdh = 0x00000000;
	*fdl = 0x7fbfffff;
      }
      else{
	*fdh = 0x7ff7ffff;
	*fdl = 0xffffffff;
      }
    }
    else if(f_e||madd_e){
      *e = 1;
      *v = 0;
      *z = 0;
      *o = 0;
      *u = 0;
      *i = 0;
    }
    else{
      ROUND(rm);
      EXCEPTION(e, v, z, o, u, i);
      if(!fp_except)
        REPACK(fdh, fdl);
      else
        SP_RES(fdh, fdl);
    }
}

void PS_FPU(int   op,               /* operation */
	 int   rm,               /* rounding mode */
         int   fs_bit,           /* FS bit    */
         ulong fsh,  ulong fsl,  /* source    */
         ulong fth,  ulong ftl,  /* target    */
         ulong *fdh, ulong *fdl, /* destination */
         int *e,   /* Unimplement Operation */
         int *v,   /* Invalid Operation     */
         int *z,   /* Divide by Zero        */
         int *o,   /* Overflow              */
         int *u,   /* Underflow             */
         int *i,   /* Inexact Operation     */
         int *cmp, /* Compare?              */
         int *cc   /* Condition Code        */
)
{
  int e0,v0,z0,o0,u0,i0,cmp0,cc0;
  int e1,v1,z1,o1,u1,i1,cmp1,cc1;
  ulong d0,d1,d;
  FPU(op,fmtS,rm,fs_bit,0,fsl,0,ftl,&d,&d0,
      &e0,&v0,&z0,&o0,&u0,&i0,&cmp0,&cc0);
  FPU(op,fmtS,rm,fs_bit,0,fsh,0,fth,&d,&d1,
      &e1,&v1,&z1,&o1,&u1,&i1,&cmp1,&cc1);
  *fdh = d1;
  *fdl = d0;
  if(e0){
    v0 = z0 = o0 = u0 = i0 = 0;
  }
  if(e1){
    v1 = z1 = o1 = u1 = i1 = 0;
  }
  *e = e1 | e0;
  *v = v1 | v0;
  *z = z1 | z0;
  *o = o1 | o0;
  *u = u1 | u0;
  *i = i1 | i0;
  *cmp = cmp1 + cmp0;
  *cc = (cc1<<1)|(cc0);
}

void PS_FPUX(int   op,               /* operation */
	 int   rm,               /* rounding mode */
         int   fs_bit,           /* FS bit    */
         ulong fsh,  ulong fsl,  /* source    */
         ulong fth,  ulong ftl,  /* target    */
	 ulong frh,  ulong frl,  /* rest?     */
         ulong *fdh, ulong *fdl, /* destination */
         int *e,   /* Unimplement Operation */
         int *v,   /* Invalid Operation     */
         int *z,   /* Divide by Zero        */
         int *o,   /* Overflow              */
         int *u,   /* Underflow             */
         int *i   /* Inexact Operation     */
)
{
  int e0,v0,z0,o0,u0,i0;
  int e1,v1,z1,o1,u1,i1;
  ulong d0,d1,d;
  FPUX(op-6,rm,fs_bit,0,fsl,0,ftl,0,frl,&d,&d0,
      &e0,&v0,&z0,&o0,&u0,&i0);
  FPUX(op-6,rm,fs_bit,0,fsh,0,fth,0,frh,&d,&d1,
      &e1,&v1,&z1,&o1,&u1,&i1);
  *fdh = d1;
  *fdl = d0;
  if(e0){
    v0 = z0 = o0 = u0 = i0 = 0;
  }
  if(e1){
    v1 = z1 = o1 = u1 = i1 = 0;
  }
  *e = e1 | e0;
  *v = v1 | v0;
  *z = z1 | z0;
  *o = o1 | o0;
  *u = u1 | u0;
  *i = i1 | i0;
}

