/*************************************************/
/* All rights reserved by Akira Nakamori. This   */
/* program must be uesed solely for the purpose  */ 
/* for which it was furnished by Akira Nakamori. */
/*                                               */
/*     Copyright (C) Akira Nakamori 2004         */
/*************************************************/
#include <stdio.h>
#include "define.h"
extern int sig_stop;
static int is_opened=0;
static FILE *inp;
/*
  Memory Module
*/
static char mem[MAXPAGENUM][0x10000];
static ulong pair[MAXPAGENUM];
static int pair_ptr= -1;
static char inpbuf[MAXINPBUF];
static int inp_ptr = -1;
static int get_chars = -1;

void mem_init(void)
{
  pair_ptr= -1;
}

void io_init(void)
{
  if(is_opened){
    fseek(inp, 0, SEEK_SET);
  }
}

static long addr_mod(long addr)
{
  if((addr&0xc0000000L)==0x80000000L)
    return addr&0x1ffffffcL;
  return addr&0xfffffffcL;
}

static int search(ulong addr)
{
  int i;

  if(pair_ptr<0){ /* init */
    pair_ptr = 0;
    pair[pair_ptr] = addr&0xffff0000L;
    return pair_ptr;
  }
  for(i=0;i<=pair_ptr;i++){
    if(((pair[i]^addr)&0xffff0000L)==0)
      break;
  }
  if(i<=pair_ptr){ /* match */
    return i;
  }
  if((pair_ptr+1)<MAXPAGENUM){ /* allocate */
    pair_ptr++;
    pair[pair_ptr] = addr&0xffff0000L;
    return pair_ptr;
  }
  printf("mem.c:out of memory at 0x%08x\n",addr);
  return -1;
}

ulong get_pword(ulong addr)
{
  int i;
  int c0,c1,c2,c3;

  addr &= 0xfffffffcL;
#if 1  /* extra I/O */
  if(addr==0x0f010020){
    int ch;
    if(inp_ptr < get_chars){
      inp_ptr++;
      return inpbuf[inp_ptr-1];
    }
    else{
      get_chars = 0;
      do {
        ch = getchar();
        if(get_chars<(MAXINPBUF-1))
          inpbuf[get_chars++] = ch;
      } while((ch!='\n')&&(ch!=EOF));
      if(get_chars>0){
        inp_ptr = 1;
        return inpbuf[0];
      }
      inp_ptr = get_chars;
    }
  }
  else if(addr==0x0f010040){
    if(is_opened)
      return fgetc(inp);
    else{
      inp = fopen(INPUTFILE,"rb");
      if(inp == NULL)
        return 0xffffffff;
      is_opened = 1;
      return fgetc(inp);
    }
  }
#endif
  i = search(addr);
  if(i>=0){
    addr &= 0xffff;
    c0 = mem[i][addr+0]&0xff;
    c1 = mem[i][addr+1]&0xff;
    c2 = mem[i][addr+2]&0xff;
    c3 = mem[i][addr+3]&0xff;
    return (c3<<24)|(c2<<16)|(c1<<8)|(c0);
  }
  return 0;
}

void put_pword(ulong addr, ulong data)
{
  int i;
  int c0,c1,c2,c3;

  addr &= 0xfffffffcL;
#if 1  /* extra I/O */
  if(addr==0x0f010000){
    putchar(data&0xff);
    return;
  }
#endif
  i = search(addr);
  if(i>=0){
    addr &= 0xffff;
    c0 = (data    )&0xff;
    c1 = (data>> 8)&0xff;
    c2 = (data>>16)&0xff;
    c3 = (data>>24)&0xff;
    mem[i][addr+0] = c0;
    mem[i][addr+1] = c1;
    mem[i][addr+2] = c2;
    mem[i][addr+3] = c3;
  }
}

ulong get_word(ulong addr)
{
  return get_pword(addr_mod(addr));
}

void put_word(ulong addr, ulong data)
{
  put_pword(addr_mod(addr), data);
}

void put_byte(ulong addr, ulong data)
{
  long dat;
  dat = get_pword(addr_mod(addr));
  data &= 0xff;
  switch(addr&3){
  case 0:
    dat = (dat & 0xffffff00L) | (data);
    break;
  case 1:
    dat = (dat & 0xffff00ffL) | (data<<8);
    break;
  case 2:
    dat = (dat & 0xff00ffffL) | (data<<16);
    break;
  case 3:
    dat = (dat & 0x00ffffffL) | (data<<24);
    break;
  }
  put_pword(addr_mod(addr), dat);
}

void put_half(ulong addr, ulong data)
{
  long dat;
  dat = get_pword(addr_mod(addr));
  data &= 0xffff;
  switch(addr&3){
  case 0:
    dat = (dat & 0xffff0000L) | (data);
    break;
  default:
    dat = (dat & 0x0000ffffL) | (data<<16);
  }
  put_pword(addr_mod(addr), dat);
}

void mem_dump(ulong start, ulong end, int unit)
{
  ulong  m_start;
  ulong  m_end;
  ulong  a,c;
  int    width,ainc;
  int    data;
  unsigned char   asc[16];
  unsigned char   d0,d1,d2,d3;

  if(unit==0){
    width=16;
    ainc =1;
  }
  else if(unit==1){
    width=8;
    ainc =2;
  }
  else {
    unit =2;
    width=4;
    ainc =4;
  }
  m_start = start&0xfffffffcL;
  m_end   = end  &0xfffffffcL;
  for(a=m_start,c=0;(a<m_end)&&(sig_stop==0);a += 4){
    if(c==0) printf("%08x: ",a);
    data=get_pword(a);
    asc[4*c+0]=d0=data&0xff;
    asc[4*c+1]=d1=(data>>8)&0xff;
    asc[4*c+2]=d2=(data>>16)&0xff;
    asc[4*c+3]=d3=(data>>24)&0xff;
    switch(unit){
    case 0:
      printf("%02x %02x %02x %02x  ",d0,d1,d2,d3);
      break;
    case 1:
      printf("%04x %04x ",(d1<<8)|d0,(d3<<8)|d2);
      break;
    case 2:
      printf("%08x ",data);
      break;
    }
    asc[4*c+0]=((d0>=' ')&&(d0<='~')) ? d0 : '.';
    asc[4*c+1]=((d1>=' ')&&(d1<='~')) ? d1 : '.';
    asc[4*c+2]=((d2>=' ')&&(d2<='~')) ? d2 : '.';
    asc[4*c+3]=((d3>=' ')&&(d3<='~')) ? d3 : '.';
    c++;
    if(c==4){
      int     i;
      c=0;
      for(i=0;i<16;i++) printf("%c",asc[i]);
      printf("\n");
    }
  }
  if(c!=0){
    int i;
    while(c<4){
      switch(unit){
      case 0:
	printf("            ",d0,d1,d2,d3);
	break;
      case 1:
	printf("          ",(d1<<8)|d0,(d3<<8)|d2);
	break;
      case 2:
	printf("         ",data);
	break;
      }
      asc[4*c+0]=' ';
      asc[4*c+1]=' ';
      asc[4*c+2]=' ';
      asc[4*c+3]=' ';
      c++;
    }
    for(i=0;i<16;i++) printf("%c",asc[i]);
    printf("\n");
  }
}
