/*************************************************/
/* All rights reserved by Akira Nakamori. This   */
/* program must be uesed solely for the purpose  */ 
/* for which it was furnished by Akira Nakamori. */
/*                                               */
/*     Copyright (C) Akira Nakamori 2004         */
/*************************************************/
#include <stdio.h>
#include "define.h"
/*
  Branch Brediction Statistics
*/
static int BT_ENTRY = 4;
static ulong bt_addr [512];
static int   bt_valid[512];
static int   bt_lru  [512];
static int   bt_hist [512];
static double bt_acc;
static double bt_hit;
static double bt_taken;

void set_bt(int entry)
{
  BT_ENTRY = entry;
  bt_init();
}

void bt_clear(void)
{
  int i;
  for(i=0;i<BT_ENTRY;i++){
    bt_valid[i] = 0;
    bt_hist [i] = 0;
    bt_lru  [i] = i;
  }
}

void bt_init(void)
{
  bt_acc = 0.0;
  bt_hit = 0.0;
  bt_taken = 0.0;
  bt_clear();
}

void bt_lru_upd(int entry)
{
  int i,j;
#if 1  /* LRU */
  for(i=0;i<BT_ENTRY;i++){
    if(entry == bt_lru[i]){
      for(j=i;j<(BT_ENTRY-1);j++)
	bt_lru[j]=bt_lru[j+1];
      bt_lru[BT_ENTRY-1]=entry;
      i = BT_ENTRY;
    }
  }
#else /* Random */
  bt_lru[0]=(bt_lru[0]+1)%BT_ENTRY;
#endif
}

void bt_mru_upd(int entry)
{
  int i,j;
  for(i=0;i<BT_ENTRY;i++){
    if(entry == bt_lru[i]){
      for(j=i;j>0;j--)
	bt_lru[j]=bt_lru[j-1];
      bt_lru[0]=entry;
      i = BT_ENTRY;
    }
  }
}

void bt_refer(ulong addr, int taken)
{
  int i;
  int bt_hit0;
  bt_acc++;
  for(i=0;i<BT_ENTRY;i++){
    if(addr==bt_addr[i])
      break;
  }
  bt_hit0 = (i<BT_ENTRY)&&bt_valid[i];
  if(bt_hit0&&taken){/* hit */
    bt_hit++;
    bt_hist[i] = 0;
    bt_lru_upd(i);
  }
  else if(!bt_hit0&&taken){ /* miss & taken */
    i = bt_lru[0];
    bt_valid[i] = 1;
    bt_addr [i] = addr;
    bt_hist[i] = 0;
    bt_lru_upd(i);
  }
  else if(!bt_hit0&&!taken){ /* miss & no taken -> hit */
    bt_hit++;
  }
  else{ /* hit & no taken */
#if 1
    if(bt_hist[i]){
      bt_hist[i] = 0;
      bt_valid[i] = 0;
#if 1
      bt_mru_upd(i);
#endif
    }
    else{
      bt_hist[i] = 1;
    }
#endif
  }
  if(taken)
    bt_taken++;
}

void bt_report(void)
{
  int i;
#if 0
  for(i=0;i<BT_ENTRY;i++)
    printf("%d: 0x%08x\n", i, bt_addr[i]);
#endif
  printf("bt hit %.0f/%.0f = %.2f %%\n", bt_hit, bt_acc, (bt_hit/bt_acc)*100.0);
  printf("taken  %.0f/%.0f = %.2f %%\n", bt_taken, bt_acc, (bt_taken/bt_acc)*100.0);
}

extern double ic_refer,dc_refer,ic_hit,dc_hit;
void ic_report(void)
{
  printf("ic hit %.0f/%.0f = %.2f %%\n",ic_hit,ic_refer,(ic_hit/ic_refer)*100.0);
}
void dc_report(void)
{
  printf("dc hit %.0f/%.0f = %.2f %%\n",dc_hit,dc_refer,(dc_hit/dc_refer)*100.0);
}

